#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "libs/buffer.h"

#include "newline-parser.h"

void
newline_parser_init(struct newline_parser *parser)
{
	assert(parser);

	parser->state = NEWLINE_PARSER_INIT;
}

void
newline_parser_deinit(struct newline_parser *parser)
{
	assert(parser);

	/* buffer is deinitialized at this point */
}

void
newline_parser_parse(struct newline_parser *parser,
	const char *s, unsigned int l, int output)
{
	unsigned int i;

	assert(parser);
	assert(s);

	for(i = 0; i < l; i++) {
		switch(parser->state) {
		case NEWLINE_PARSER_INIT:
			buffer_init(&parser->buffer);

			if(s[i] != '\n') {
				if(output)
					buffer_write_char(&parser->buffer,
						s[i]);
			}

			parser->state = NEWLINE_PARSER_TEXT;
			break;
		case NEWLINE_PARSER_TEXT:
			if(s[i] == '\n') {
				if(parser->buffer.i)
					parser->on_text(parser,
						parser->buffer.data,
						parser->buffer.i);
				buffer_deinit(&parser->buffer, 1);

				parser->state = NEWLINE_PARSER_NEWLINE_FIRST;
				break;
			}

			if(output)
				buffer_write_char(&parser->buffer, s[i]);

			break;
		case NEWLINE_PARSER_NEWLINE_FIRST:
			/* Slurp spaces that follow a newline */
			if(s[i] == ' ' || s[i] == '\t') {
				parser->state = NEWLINE_PARSER_NEWLINE_MANY;
				break;
			}

			if(s[i] != '\n') {
				buffer_init(&parser->buffer);

				/* Add a space since the first line
				 * is unlikely to begin with one. If
				 * there's already a space there, no
				 * harm is done. Also add it regardless of
				 * whether we should output or not. */
				buffer_write_char(&parser->buffer, ' ');

				if(output)
					buffer_write_char(&parser->buffer, s[i]);

				parser->state = NEWLINE_PARSER_TEXT;
				break;
			}

			parser->state = NEWLINE_PARSER_NEWLINE_MANY;
			break;
		case NEWLINE_PARSER_NEWLINE_MANY:
			if(s[i] == ' ' || s[i] == '\t')
				break;

			if(s[i] != '\n') {
				parser->on_newline(parser);
				buffer_init(&parser->buffer);

				if(output)
					buffer_write_char(&parser->buffer, s[i]);

				parser->state = NEWLINE_PARSER_TEXT;
				break;
			}

			break;
		}
	}
}

/* This is a "soft" finish. Call the callbacks, but keep the parser and the
 * parser state initialized. */
void
newline_parser_flush(struct newline_parser *parser)
{
	switch(parser->state) {
	case NEWLINE_PARSER_INIT:
		break;
	case NEWLINE_PARSER_TEXT:
		if(!parser->buffer.i)
			break;

		parser->on_text(parser,
			parser->buffer.data, parser->buffer.i);
		buffer_deinit(&parser->buffer, 1);
		buffer_init(&parser->buffer);
		break;
	case NEWLINE_PARSER_NEWLINE_FIRST:
		break;
	case NEWLINE_PARSER_NEWLINE_MANY:
		break;
	}
}

void
newline_parser_finish(struct newline_parser *parser)
{
	switch(parser->state) {
	case NEWLINE_PARSER_INIT:
		/* buffer is not even initialized. */
		break;
	case NEWLINE_PARSER_TEXT:
		if(!parser->buffer.i)
			break;

		parser->on_text(parser,
			parser->buffer.data, parser->buffer.i);
		buffer_deinit(&parser->buffer, 1);
		break;
	case NEWLINE_PARSER_NEWLINE_FIRST:
		/* buffer is not even initialized. */
		break;
	case NEWLINE_PARSER_NEWLINE_MANY:
		/* buffer is not even initialized. */
		break;
	}
}
