#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libxml/parser.h>

#include "libs/buffer.h"
#include "libs/list.h"
#include "libs/xmalloc.h"
#include "libs/xml.h"

#include "request-parser.h"

/* TextForExplanation XML request parser; see TextForExplanation.xsd */

void
request_parser_init(struct request_parser *parser)
{
	parser->state = REQUEST_PARSER_ROOT;
	parser->error = NULL;
}

void
request_parser_deinit(struct request_parser *parser)
{
	free(parser->error);
}

static void
st_startElement(void *data, const xmlChar *name, const xmlChar **attrs)
{
	struct request_parser *parser;
	const struct request_parser_handler *handler;

	assert(data);

	parser = data;
	handler = parser->handler;

	switch(parser->state) {
	case REQUEST_PARSER_ROOT:
		if(!strcmp((char *) name, "TextForExplanation")) {
			handler->on_start_textforexplanation(parser->request);
			parser->state = REQUEST_PARSER_TEXTFOREXPLANATION;
			break;
		}

		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_TEXTFOREXPLANATION:
		if(!strcmp((char *) name, "DocCategory")) {
			handler->on_start_doccategory(parser->request);
			parser->state = REQUEST_PARSER_DOCCATEGORY;
			break;
		}

		if(!strcmp((char *) name, "ElementCategory")) {
			handler->on_start_elementcategory(parser->request);
			parser->state = REQUEST_PARSER_ELEMENTCATEGORY;
			break;
		}

		if(!strcmp((char *) name, "SiteUnit")) {
			handler->on_start_siteunit(parser->request);
			parser->state = REQUEST_PARSER_SITEUNIT;
			break;
		}

		if(!strcmp((char *) name, "Text")) {
			handler->on_start_text(parser->request);
			parser->state = REQUEST_PARSER_TEXT;
			break;
		}

		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_DOCCATEGORY:
		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_ELEMENTCATEGORY:
		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_SITEUNIT:
		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_TEXT:
		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_ERROR:
		return;
	case REQUEST_PARSER_DONE:
		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		return;
	}
}

static void
st_endElement(void *data, const xmlChar *name)
{
	struct request_parser *parser;
	const struct request_parser_handler *handler;

	assert(data);

	parser = data;
	handler = parser->handler;

	switch(parser->state) {
	case REQUEST_PARSER_ROOT:
		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_TEXTFOREXPLANATION:
		if(!strcmp((char *) name, "TextForExplanation")) {
			handler->on_end_textforexplanation(parser->request);
			parser->state = REQUEST_PARSER_DONE;
			break;
		}

		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_DOCCATEGORY:
		if(!strcmp((char *) name, "DocCategory")) {
			handler->on_end_doccategory(parser->request);
			parser->state = REQUEST_PARSER_TEXTFOREXPLANATION;
			break;
		}

		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_ELEMENTCATEGORY:
		if(!strcmp((char *) name, "ElementCategory")) {
			handler->on_end_elementcategory(parser->request);
			parser->state = REQUEST_PARSER_TEXTFOREXPLANATION;
			break;
		}

		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_SITEUNIT:
		if(!strcmp((char *) name, "SiteUnit")) {
			handler->on_end_siteunit(parser->request);
			parser->state = REQUEST_PARSER_TEXTFOREXPLANATION;
			break;
		}

		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_TEXT:
		if(!strcmp((char *) name, "Text")) {
			handler->on_end_text(parser->request);
			parser->state = REQUEST_PARSER_TEXTFOREXPLANATION;
			break;
		}

		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_ERROR:
		return;
	case REQUEST_PARSER_DONE:
		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		return;
	}
}

static void
st_characters(void *data, const xmlChar *ch, int len)
{
	struct request_parser *parser;
	const struct request_parser_handler *handler;

	assert(data);

	parser = data;
	handler = parser->handler;

	switch(parser->state) {
	case REQUEST_PARSER_ROOT:
		/* ignore */
		break;
	case REQUEST_PARSER_TEXTFOREXPLANATION:
		/* ignore */
		break;
	case REQUEST_PARSER_DOCCATEGORY:
		handler->on_characters_doccategory(parser->request,
			(const char *) ch, len);
		break;
	case REQUEST_PARSER_ELEMENTCATEGORY:
		handler->on_characters_elementcategory(parser->request,
			(const char *) ch, len);
		break;
	case REQUEST_PARSER_SITEUNIT:
		handler->on_characters_siteunit(parser->request,
			(const char *) ch, len);
		break;
	case REQUEST_PARSER_TEXT:
		handler->on_characters_text(parser->request,
			(const char *) ch, len);
		break;
	case REQUEST_PARSER_ERROR:
		return;
	case REQUEST_PARSER_DONE:
		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		return;
	}
}

static char *
st_vstrdup(const char *msg, va_list ap)
{
	char buf[1024];

	vsnprintf(buf, sizeof(buf), msg, ap);

	return strdup(buf);
}

static void
st_warning(void *data, const char *msg, ...)
{
	struct request_parser *parser;
	va_list ap;

	assert(data);

	parser = data;
}

static void
st_error(void *data, const char *msg, ...)
{
	struct request_parser *parser;
	va_list ap;

	assert(data);

	parser = data;
	parser->state = REQUEST_PARSER_ERROR;

	va_start(ap, msg);
	parser->error = st_vstrdup(msg, ap);
	va_end(ap);
}

static void
st_fatalError(void *data, const char *msg, ...)
{
	struct request_parser *parser;
	va_list ap;

	assert(data);

	parser = data;
	parser->state = REQUEST_PARSER_ERROR;

	va_start(ap, msg);
	parser->error = st_vstrdup(msg, ap);
	va_end(ap);
}

const xmlSAXHandler request_parser_handler = {
	.startElement = &st_startElement,
	.endElement = &st_endElement,
	.characters = &st_characters,
	.warning = &st_warning,
	.error = &st_error,
	.fatalError = &st_fatalError,
};
