#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libxml/parser.h>

#include "libs/buffer.h"
#include "libs/list.h"
#include "libs/xmalloc.h"
#include "libs/xml.h"
#include "libs/xsd.h"

#include "response-parser.h"

/* thesaurus response parser; see response.xsd */

void
response_parser_init(struct response_parser *parser)
{
	parser->state = RESPONSE_PARSER_ROOT;
	parser->error = NULL;
}

void
response_parser_deinit(struct response_parser *parser)
{
	free(parser->error);
}

static void
st_startElement(void *data, const xmlChar *name, const xmlChar **attrs)
{
	struct response_parser *parser;
	const struct response_parser_handler *handler;

	assert(data);

	parser = data;
	handler = parser->handler;

	switch(parser->state) {
	case RESPONSE_PARSER_ROOT:
		if(!strcmp((char *) name, "document")) {
			handler->on_start_document(parser->response);
			parser->state = RESPONSE_PARSER_DOCUMENT;
			break;
		}

		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		break;
	case RESPONSE_PARSER_DOCUMENT:
		if(!strcmp((char *) name, "text")) {
			handler->on_start_text(parser->response);
			parser->state = RESPONSE_PARSER_TEXT;
			break;
		}

		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		break;
	case RESPONSE_PARSER_TEXT:
		if(!strcmp((char *) name, "word")) {
			handler->on_start_word(parser->response);
			parser->state = RESPONSE_PARSER_WORD;
			break;
		}

		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		break;
	case RESPONSE_PARSER_WORD:
		if(!strcmp((char *) name, "original")) {
			handler->on_start_original(parser->response);
			parser->state = RESPONSE_PARSER_ORIGINAL;
			break;
		}

		if(!strcmp((char *) name, "result")) {
			int i;

			int exact = 0;
			int stemmed = 0;
			int translated = 0;
			int edit_distance = 0;

			for(i = 0; attrs[i]; i += 2) {
				if(!strcmp((char *) attrs[i], "exact")) {
					exact = xsd_boolean((char *) attrs[i + 1]);
					continue;
				}

				if(!strcmp((char *) attrs[i], "stemmed")) {
					stemmed = xsd_boolean((char *) attrs[i + 1]);
					continue;
				}

				if(!strcmp((char *) attrs[i], "translated")) {
					translated = xsd_boolean((char *) attrs[i + 1]);
					continue;
				}

				if(!strcmp((char *) attrs[i], "editDistance")) {
					edit_distance = xsd_positive_integer((char *) attrs[i + 1]);
					continue;
				}
			}

			handler->on_start_result(parser->response,
				exact, stemmed, translated, edit_distance);
			parser->state = RESPONSE_PARSER_RESULT;
			break;
		}

		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		break;
	case RESPONSE_PARSER_ORIGINAL:
		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		break;
	case RESPONSE_PARSER_RESULT:
		if(!strcmp((char *) name, "synonym")) {
			handler->on_start_synonym(parser->response);
			parser->state = RESPONSE_PARSER_SYNONYM;
			break;
		}

		if(!strcmp((char *) name, "term")) {
			handler->on_start_term(parser->response);
			parser->state = RESPONSE_PARSER_TERM;
			break;
		}

		if(!strcmp((char *) name, "explanation")) {
			handler->on_start_explanation(parser->response);
			parser->state
				= RESPONSE_PARSER_EXPLANATION;
			break;
		}

		if(!strcmp((char *) name, "thesaurus")) {
			handler->on_start_thesaurus(parser->response);
			parser->state
				= RESPONSE_PARSER_THESAURUS;
			break;
		}

		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		break;
	case RESPONSE_PARSER_SYNONYM:
		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		break;
	case RESPONSE_PARSER_TERM:
		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		break;
	case RESPONSE_PARSER_EXPLANATION:
		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		break;
	case RESPONSE_PARSER_THESAURUS:
		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		break;
	case RESPONSE_PARSER_ERROR:
		return;
	case RESPONSE_PARSER_DONE:
		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		return;
	}
}

static void
st_endElement(void *data, const xmlChar *name)
{
	struct response_parser *parser;
	const struct response_parser_handler *handler;

	assert(data);

	parser = data;
	handler = parser->handler;

	switch(parser->state) {
	case RESPONSE_PARSER_ROOT:
		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		break;
	case RESPONSE_PARSER_DOCUMENT:
		if(!strcmp((char *) name, "document")) {
			handler->on_end_document(parser->response);
			parser->state = RESPONSE_PARSER_DONE;
			break;
		}

		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		break;
	case RESPONSE_PARSER_TEXT:
		if(!strcmp((char *) name, "text")) {
			handler->on_end_text(parser->response);
			parser->state = RESPONSE_PARSER_DOCUMENT;
			break;
		}

		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		break;
	case RESPONSE_PARSER_WORD:
		if(!strcmp((char *) name, "word")) {
			handler->on_end_word(parser->response);
			parser->state = RESPONSE_PARSER_TEXT;
			break;
		}

		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		break;
	case RESPONSE_PARSER_ORIGINAL:
		if(!strcmp((char *) name, "original")) {
			handler->on_end_original(parser->response);
			parser->state = RESPONSE_PARSER_WORD;
			break;
		}

		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		break;
	case RESPONSE_PARSER_RESULT:
		if(!strcmp((char *) name, "result")) {
			handler->on_end_result(parser->response);
			parser->state = RESPONSE_PARSER_WORD;
			break;
		}

		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		break;
	case RESPONSE_PARSER_SYNONYM:
		if(!strcmp((char *) name, "synonym")) {
			handler->on_end_synonym(parser->response);
			parser->state = RESPONSE_PARSER_RESULT;
			break;
		}

		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		break;
	case RESPONSE_PARSER_TERM:
		if(!strcmp((char *) name, "term")) {
			handler->on_end_term(parser->response);
			parser->state = RESPONSE_PARSER_RESULT;
			break;
		}

		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		break;
	case RESPONSE_PARSER_EXPLANATION:
		if(!strcmp((char *) name, "explanation")) {
			handler->on_end_explanation(parser->response);
			parser->state = RESPONSE_PARSER_RESULT;
			break;
		}

		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		break;
	case RESPONSE_PARSER_THESAURUS:
		if(!strcmp((char *) name, "thesaurus")) {
			handler->on_end_thesaurus(parser->response);
			parser->state = RESPONSE_PARSER_RESULT;
			break;
		}

		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		break;
	case RESPONSE_PARSER_ERROR:
		return;
	case RESPONSE_PARSER_DONE:
		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		return;
	}
}

static void
st_characters(void *data, const xmlChar *ch, int len)
{
	struct response_parser *parser;
	const struct response_parser_handler *handler;

	assert(data);

	parser = data;
	handler = parser->handler;

	switch(parser->state) {
	case RESPONSE_PARSER_ROOT:
		/* ignore */
		break;
	case RESPONSE_PARSER_DOCUMENT:
		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		break;
	case RESPONSE_PARSER_TEXT:
		handler->on_characters_text(parser->response,
			(const char *) ch, len);
		break;
	case RESPONSE_PARSER_WORD:
		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		break;
	case RESPONSE_PARSER_ORIGINAL:
		handler->on_characters_original(parser->response,
			(const char *) ch, len);
		break;
	case RESPONSE_PARSER_RESULT:
		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		break;
	case RESPONSE_PARSER_SYNONYM:
		handler->on_characters_synonym(parser->response,
			(const char *) ch, len);
		break;
	case RESPONSE_PARSER_TERM:
		handler->on_characters_term(parser->response,
			(const char *) ch, len);
		break;
	case RESPONSE_PARSER_EXPLANATION:
		handler->on_characters_explanation(parser->response,
			(const char *) ch, len);
		break;
	case RESPONSE_PARSER_THESAURUS:
		handler->on_characters_thesaurus(parser->response,
			(const char *) ch, len);
		break;
	case RESPONSE_PARSER_ERROR:
		return;
	case RESPONSE_PARSER_DONE:
		/* error */
		parser->state = RESPONSE_PARSER_ERROR;
		return;
	}
}

static char *
st_vstrdup(const char *msg, va_list ap)
{
	char buf[1024];

	vsnprintf(buf, sizeof(buf), msg, ap);

	return strdup(buf);
}

static void
st_warning(void *data, const char *msg, ...)
{
	struct response_parser *parser;

	assert(data);

	parser = data;
}

static void
st_error(void *data, const char *msg, ...)
{
	struct response_parser *parser;
	va_list ap;

	assert(data);

	parser = data;
	parser->state = RESPONSE_PARSER_ERROR;

	va_start(ap, msg);
	parser->error = st_vstrdup(msg, ap);
	va_end(ap);
}

static void
st_fatalError(void *data, const char *msg, ...)
{
	struct response_parser *parser;
	va_list ap;

	assert(data);

	parser = data;
	parser->state = RESPONSE_PARSER_ERROR;

	va_start(ap, msg);
	parser->error = st_vstrdup(msg, ap);
	va_end(ap);
}

const xmlSAXHandler response_parser_handler = {
	.startElement = &st_startElement,
	.endElement = &st_endElement,
	.characters = &st_characters,
	.warning = &st_warning,
	.error = &st_error,
	.fatalError = &st_fatalError,
};
