#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libxml/parser.h>

#include "libs/buffer.h"
#include "libs/list.h"
#include "libs/uri.h"
#include "libs/xmalloc.h"
#include "libs/xml.h"

#include "response.h"
#include "response-parser.h"

void
response_init(struct response *response)
{
	assert(response);

	buffer_init(&response->output);
	response->explanation_id = 0;
}

void
response_deinit(struct response *response)
{
	assert(response);

	buffer_deinit(&response->output, 1);
}

void
response_word_init(struct response_word *word)
{
	assert(word);

	word->original = NULL;
	word->original_length = 0;

	LIST_INIT(&word->results);
}

void
response_word_deinit(struct response_word *word)
{
	struct response_result *result;
	struct response_result *result_temp;

	assert(word);

	free(word->original);

	LIST_FOR_SAFE(&word->results, result, result_temp) {
		response_result_deinit(result);
		free(result);
	}
}

void
response_result_init(struct response_result *result)
{
	LIST_INIT(result);

	result->exact = 0;
	result->stemmed = 0;
	result->translated = 0;
	result->edit_distance = 0;

	result->synonym = NULL;
	result->synonym_length = 0;

	result->term = NULL;
	result->term_length = 0;

	result->explanation = NULL;
	result->explanation_length = 0;

	result->thesaurus = NULL;
	result->thesaurus_length = 0;
}

void
response_result_deinit(struct response_result *result)
{
	free(result->synonym);
	free(result->term);
	free(result->explanation);
	free(result->thesaurus);
}

static void
on_start_document(struct response *response)
{
	struct buffer *output;

	output = &response->output;
	buffer_write_string(output,
		"<?xml "
		"version=\"1.0\" "
		"encoding=\"UTF-8\"?>");
	buffer_write_string(output,
		"<!DOCTYPE html PUBLIC "
		"\"-//W3C//DTD XHTML 1.0 Strict//NN\" "
		"\"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">");
	buffer_write_string(output,
		"<html "
		"xmlns=\"http://www.w3.org/1999/xhtml\" "
		"xml:lang=\"en\" "
		"lang=\"en\">");
	buffer_write_string(output, "<head>");
	buffer_write_format(output, "<title>%s</title>",
		"MinJournal2 Thesaurus (Viewer)");
	buffer_write_string(output,
		"<link "
		"rel=\"stylesheet\" "
		"type=\"text/css\" "
		"href=\"default.css\" />");
	buffer_write_string(output,
		"<script "
		"src=\"thesaurus.js\" "
		"type=\"text/javascript\" />");
	buffer_write_string(output, "</head>");
	buffer_write_string(output, "<body>");
}

static void
on_end_document(struct response *response)
{
	struct buffer *output;

	output = &response->output;
	buffer_write_string(output, "</body>");
	buffer_write_string(output, "</html>");
}

static void
st_on_text(struct newline_parser *parser, const char *s, unsigned int l)
{
	struct response *response;

	response = parser->data;

	xml_escape_to_buffer(s, l, &response->output);
}

static void
st_on_newline(struct newline_parser *parser)
{
	struct response *response;

	response = parser->data;

	buffer_write_string(&response->output, "<br />");
	buffer_write_string(&response->output, "<br />");
}

static void
on_start_text(struct response *response)
{
	/* begin text reading */

	newline_parser_init(&response->text);
	response->text.data = response;
	response->text.on_text = &st_on_text;
	response->text.on_newline = &st_on_newline;

}

static void
on_characters_text(struct response *response, const char *s, unsigned int l)
{
	newline_parser_parse(&response->text, s, l, 1);
}

static void
on_end_text(struct response *response)
{
	/* end text reading */
	newline_parser_finish(&response->text);
	newline_parser_deinit(&response->text);
}

static void
on_start_word(struct response *response)
{
	struct response_word *word;

	word = xmalloc(sizeof(*word));
	response_word_init(word);
	response->word = word;
}

/* Control whether the mouse must be clicked on a term in order to display
 * the explanation, or whether simply moving the mouse over the term will
 * show the explanation. */
#define CONFIG_CLICK

/* Print brackets before and after all the explanations. */
#define CONFIG_BRACKETS

/* Start each explanation on a new line. */
#undef CONFIG_LINEBREAKS

static void
on_word_result_patient(struct response *response)
{
	struct buffer *output;
	struct response_word *word;
	struct response_result *result;

	output = &response->output;
	word = response->word;

	if(LIST_IS_EMPTY(&word->results)) {
		/* No results, just print out the word. */
		xml_escape_to_buffer(word->original,
			word->original_length, output);
		return;
	}

	/* The word was deemed a stopword. Don't print any explanations. */
	LIST_FOR(&word->results, result) {
		if(!result->explanation) {
			xml_escape_to_buffer(word->original,
				word->original_length, output);
			return;
		}
	}

#ifdef CONFIG_CLICK
	/* Use onClick hooks */
	buffer_write_format(output,
		"<span "
		"class=\"word\" "
		"onClick=\"toggle(\'ex_%d\')\">",
		response->explanation_id);
#else
	/* Use onMouseOver/onMouseOut hooks */
	buffer_write_format(output,
		"<span "
		"class=\"word\" "
		"onMouseOver=\"toggle(\'ex_%d\')\" "
		"onMouseOut=\"toggle(\'ex_%d\')\">",
		response->explanation_id,
		response->explanation_id);
#endif
	xml_escape_to_buffer(word->original, word->original_length, output);
	buffer_write_string(output, "</span>");

	buffer_write_format(output,
		"<span "
		"id=\"ex_%d\" "
		"class=\"explanation\">",
		response->explanation_id);
#ifdef CONFIG_BRACKETS
	buffer_write_string(output, " [");
#endif

	LIST_FOR(&word->results, result) {
		buffer_write_format(output,
			"<span "
			"class=\"term-%s\">",
			result->exact ? "exact" : "inexact");
		xml_escape_to_buffer(result->term,
			result->term_length, output);
		buffer_write_string(output, "</span>");

		buffer_write_string(output, ": ");
		xml_escape_to_buffer(result->explanation,
			result->explanation_length, output);

#ifdef CONFIG_LINEBREAKS
		if(result != LIST_LAST(&word->results))
			buffer_write_string(output, "; <br />");
#else
		if(result != LIST_LAST(&word->results))
			buffer_write_string(output, "; ");
#endif
	}

#ifdef CONFIG_BRACKETS
	buffer_write_string(output, "] ");
#endif
	buffer_write_string(output, "</span>");

	response->explanation_id++;
}

static void
on_word_result_physician(struct response *response)
{
	struct buffer *output;
	struct response_word *word;
	struct response_result *result;

	output = &response->output;
	word = response->word;

	if(LIST_IS_EMPTY(&word->results)) {
		char *uri;
		unsigned int uri_length;

		uri_encode(word->original, word->original_length,
			&uri, &uri_length);

		/* No results. */
		buffer_write_format(output,
			"<a "
			"class=\"word-unknown\" "
			"href=\"edit?word=%.*s\" "
			"target=\"_self\">",
			uri_length, uri);
		xml_escape_to_buffer(word->original, word->original_length,
			output);
		buffer_write_string(output,
			"</a>");

		free(uri);
		return;
	}

	/* Stopword. */
	LIST_FOR(&word->results, result) {
		if(!result->explanation) {
			char *uri;
			unsigned int uri_length;

			uri_encode(word->original, word->original_length,
				&uri, &uri_length);

			buffer_write_format(output,
				"<a "
				"class=\"word-stopword\" "
				"href=\"edit?word=%.*s\" "
				"target=\"_self\">",
				uri_length, uri);
			xml_escape_to_buffer(word->original,
				word->original_length, output);
			buffer_write_string(output,
				"</a>");

			free(uri);
			return;
		}
	}

	buffer_write_format(output,
		"<span "
		"class=\"word-known-%s\" "
		"onClick=\"toggle(\'ex_%d\')\">",
		LIST_FIRST(&word->results)->exact ? "exact" : "inexact",
		response->explanation_id);
	xml_escape_to_buffer(word->original, word->original_length, output);
	buffer_write_string(output, "</span>");

	buffer_write_format(output,
		"<span "
		"id=\"ex_%d\" "
		"class=\"explanation\">",
		response->explanation_id);
	buffer_write_string(output, " [");

	LIST_FOR(&word->results, result) {
		char *uri;
		unsigned int uri_length;

		uri_encode(word->original, word->original_length,
			&uri, &uri_length);

		buffer_write_format(output,
			"<a "
			"class=\"term-%s\" "
			"href=\"edit?word=%.*s\" "
			"target=\"_self\">",
			result->exact ? "exact" : "inexact",
			uri_length, uri);
		xml_escape_to_buffer(result->term,
			result->term_length, output);
		buffer_write_string(output, "</a>");

		buffer_write_string(output, ": ");
		xml_escape_to_buffer(result->explanation,
			result->explanation_length, output);

		if(result != LIST_LAST(&word->results))
			buffer_write_string(output, "; ");

		free(uri);
	}

	buffer_write_string(output, "] ");
	buffer_write_string(output, "</span>");

	response->explanation_id++;
}

static void
on_end_word(struct response *response)
{
#if 1
	on_word_result_patient(response);
#else
	on_word_result_physician(response);
#endif

	response_word_deinit(response->word);
	free(response->word);
}

static void
on_start_original(struct response *response)
{
	/* Make sure the callbacks are called at this point. */
	newline_parser_flush(&response->text);

	/* begin original reading */
	buffer_init(&response->original);
}

static void
on_characters_original(struct response *response,
	const char *s, unsigned int l)
{
	/* Call the newline parser, but don't let it save the actual text of
	 * the word, just act as if it was there. This is necessary so that
	 * the parser can make assumptions about what the output looks like.
	 * The text will be output anyway, but not through the newline
	 * parser. */
	newline_parser_parse(&response->text, s, l, 0);

	buffer_write(&response->original, s, l);
}

static void
on_end_original(struct response *response)
{
	struct response_word *word;

	/* Make sure the callbacks are called at this point. */
	newline_parser_flush(&response->text);

	word = response->word;
	buffer_move(&response->original,
		&word->original, &word->original_length);

	/* end original reading */
	buffer_deinit(&response->original, 1);
}

static void
on_start_result(struct response *response,
	int exact, int stemmed, int translated, unsigned int edit_distance)
{
	struct response_result *result;

	result = xmalloc(sizeof(*result));
	response_result_init(result);

	result->exact = exact;
	result->stemmed = stemmed;
	result->translated = translated;
	result->edit_distance = edit_distance;

	result->synonym = NULL;
	result->synonym_length = 0;

	result->term = NULL;
	result->term_length = 0;

	result->explanation = NULL;
	result->explanation_length = 0;

	result->thesaurus = NULL;
	result->thesaurus_length = 0;

	response->result = result;
}

static void
on_end_result(struct response *response)
{
	LIST_INSERT_BEFORE(&response->word->results, response->result);
}

static void
on_start_synonym(struct response *response)
{
	/* begin synonym reading */
	buffer_init(&response->synonym);
}

static void
on_characters_synonym(struct response *response, const char *s, unsigned int l)
{
	buffer_write(&response->synonym, s, l);
}

static void
on_end_synonym(struct response *response)
{
	struct response_result *result;

	result = response->result;
	buffer_move(&response->synonym,
		&result->synonym, &result->synonym_length);

	/* end synonym reading */
	buffer_deinit(&response->synonym, 1);
}

static void
on_start_term(struct response *response)
{
	/* begin term reading */
	buffer_init(&response->term);
}

static void
on_characters_term(struct response *response, const char *s, unsigned int l)
{
	buffer_write(&response->term, s, l);
}

static void
on_end_term(struct response *response)
{
	struct response_result *result;

	result = response->result;
	buffer_move(&response->term,
		&result->term, &result->term_length);

	/* end term reading */
	buffer_deinit(&response->term, 1);
}

static void
on_start_explanation(struct response *response)
{
	/* begin explanation reading */
	buffer_init(&response->explanation);
}

static void
on_characters_explanation(struct response *response,
	const char *s, unsigned int l)
{
	buffer_write(&response->explanation, s, l);
}

static void
on_end_explanation(struct response *response)
{
	struct response_result *result;

	result = response->result;
	buffer_move(&response->explanation,
		&result->explanation, &result->explanation_length);

	/* end explanation reading */
	buffer_deinit(&response->explanation, 1);
}

static void
on_start_thesaurus(struct response *response)
{
	/* begin thesaurus reading */
	buffer_init(&response->thesaurus);
}

static void
on_characters_thesaurus(struct response *response,
	const char *s, unsigned int l)
{
	buffer_write(&response->thesaurus, s, l);
}

static void
on_end_thesaurus(struct response *response)
{
	struct response_result *result;

	result = response->result;
	buffer_move(&response->thesaurus,
		&result->thesaurus, &result->thesaurus_length);

	/* end thesaurus reading */
	buffer_deinit(&response->thesaurus, 1);
}

const struct response_parser_handler response_handler = {
	.on_start_document		= &on_start_document,
	.on_start_text			= &on_start_text,
	.on_start_word			= &on_start_word,
	.on_start_original		= &on_start_original,
	.on_start_result		= &on_start_result,
	.on_start_synonym		= &on_start_synonym,
	.on_start_term			= &on_start_term,
	.on_start_explanation		= &on_start_explanation,
	.on_start_thesaurus		= &on_start_thesaurus,

	.on_end_document		= &on_end_document,
	.on_end_text			= &on_end_text,
	.on_end_word			= &on_end_word,
	.on_end_original		= &on_end_original,
	.on_end_result			= &on_end_result,
	.on_end_synonym			= &on_end_synonym,
	.on_end_term			= &on_end_term,
	.on_end_explanation		= &on_end_explanation,
	.on_end_thesaurus		= &on_end_thesaurus,

	.on_characters_text		= &on_characters_text,
	.on_characters_original		= &on_characters_original,
	.on_characters_synonym		= &on_characters_synonym,
	.on_characters_term		= &on_characters_term,
	.on_characters_explanation	= &on_characters_explanation,
	.on_characters_thesaurus	= &on_characters_thesaurus,
};
