#include <netinet/in.h>

#include <assert.h>
#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <libxml/parser.h>

#include "libs/io-loop.h"
#include "libs/io-socket.h"
#include "libs/io-socket-buffer.h"

#include "request.h"
#include "request-parser.h"
#include "response.h"
#include "response-parser.h"
#include "text-parser.h"

/* Thesaurus connection */
static struct {
	struct io_socket socket;
	int running;

	/* XML Parser */
	xmlParserCtxtPtr response_xml_parser;
	struct response_parser response_parser;
	struct response response;
} thesaurus;

/* HTTP input stream */
static struct {
	struct io_socket socket;
	int running;

	unsigned int cgi;
	unsigned int content_length;
	unsigned int bytes_received;

	xmlParserCtxtPtr request_xml_parser;
	struct request_parser request_parser;
	struct request request;
} input;

/* HTTP output stream */
static struct {
	struct io_socket socket;
	int running;
} output;

static void
st_thesaurus_on_read(struct io_socket *socket)
{
	unsigned long int can_read;
	unsigned char *buf;

	can_read = io_socket_buffer_can_read(&socket->ibuffer);
	io_socket_buffer_read(&socket->ibuffer, &buf);

#if 0
	printf("read thesaurus:\n");
	printf("\x1b[31m[\x1b[32m%.*s\x1b[31m]\n", (int) can_read, buf);
#endif

	xmlParseChunk(thesaurus.response_xml_parser,
		(const char *) buf, can_read, 0);
	io_socket_buffer_advance(&socket->ibuffer, can_read);
	io_socket_buffer_write_buffer(&output.socket.obuffer, &thesaurus.response.output);

	if(thesaurus.response_parser.state == RESPONSE_PARSER_ERROR) {
		/* Error */
		printf("[Error Occurred While Processing Response: %d: %s\n",
			thesaurus.response_xml_parser->input->line,
			thesaurus.response_parser.error ? thesaurus.response_parser.error : "(unknown)");
		socket->close = 1;
		return;
	}

	if(thesaurus.response_parser.state == RESPONSE_PARSER_DONE) {
		socket->close = 1;
		return;
	}
}

static void
st_thesaurus_on_write(struct io_socket *socket)
{
}

static void
st_thesaurus_on_close(struct io_socket *socket)
{
	thesaurus.running = 0;
	/* XXX: Does this *really* work? */
	output.socket.close = 1;

	io_socket_deinit(socket);
}

static void
st_input_on_read(struct io_socket *socket)
{
	unsigned long int can_read;
	unsigned char *buf;

	can_read = io_socket_buffer_can_read(&socket->ibuffer);
	io_socket_buffer_read(&socket->ibuffer, &buf);

#if 0
	fprintf(stderr, "\x1b[31m[\x1b[33m%.*s\x1b[31m]\n", can_read, buf);
#endif

	xmlParseChunk(input.request_xml_parser,
		(const char *) buf, can_read, 0);
	io_socket_buffer_advance(&socket->ibuffer, can_read);
	io_socket_buffer_write_buffer(&thesaurus.socket.obuffer, &input.request.output);

	if(input.request_parser.state == REQUEST_PARSER_ERROR) {
		/* Error */
		printf("[Error Occurred While Processing Request: %d: %s]\n",
			input.request_xml_parser->input->line,
			input.request_parser.error ? input.request_parser.error : "(unknown)");
		socket->close = 1;
		return;
	}

	if(input.request_parser.state == REQUEST_PARSER_DONE) {
		socket->close = 1;
		return;
	}

	/* If we're running in CGI mode, count the number of bytes
	 * received. */
	if(input.cgi)
		input.bytes_received += can_read;
}

static void
st_input_on_close(struct io_socket *socket)
{
	input.running = 0;

	io_socket_deinit(socket);
}

static void
st_output_on_write(struct io_socket *socket)
{
}

static void
st_output_on_close(struct io_socket *socket)
{
	output.running = 0;

	io_socket_deinit(socket);
}

static void
st_cgi_init()
{
	const char *cgi;
	const char *content_length;

	input.cgi = 0;

	cgi = getenv("GATEWAY_INTERFACE");
	if(cgi && !strncmp(cgi, "CGI", 3)) {
		input.cgi = 1;

		/* Only print this if we're running as a CGI */
		fprintf(stdout, "Content-type: text/plain\n");
		fprintf(stdout, "\n");
		fflush(stdout);

		content_length = getenv("CONTENT_LENGTH");
		if(content_length) {
			input.content_length = atoi(content_length);
			input.bytes_received = 0;
		} else {
			input.content_length = 0;
			input.bytes_received = 0;
		}
	}
}

int
main(int argc, char *argv[])
{
	struct io_loop loop;

	/* Check that the libxml version in use is compatible with the
	 * version the software has been compiled against. */
	LIBXML_TEST_VERSION

	setlocale(LC_ALL, "");

	io_loop_init(&loop);

	io_socket_init(&thesaurus.socket);
	thesaurus.socket.on_read = &st_thesaurus_on_read;
	thesaurus.socket.on_write = &st_thesaurus_on_write;
	thesaurus.socket.on_close = &st_thesaurus_on_close;

	thesaurus.response_xml_parser =
		xmlCreatePushParserCtxt(&response_parser_handler,
			&thesaurus.response_parser, NULL, 0, NULL);

	response_parser_init(&thesaurus.response_parser);
	thesaurus.response_parser.response = &thesaurus.response;
	thesaurus.response_parser.handler = &response_handler;

	response_init(&thesaurus.response);

	io_socket_init(&input.socket);
	input.socket.fd = STDIN_FILENO;
	input.socket.dont_write = 1;
	input.socket.on_read = &st_input_on_read;
	input.socket.on_close = &st_input_on_close;
	input.running = 1;

	input.request_xml_parser =
		xmlCreatePushParserCtxt(&request_parser_handler,
			&input.request_parser, NULL, 0, NULL);

	request_parser_init(&input.request_parser);
	input.request_parser.request = &input.request;
	input.request_parser.handler = &request_handler;

	request_init(&input.request);

	io_socket_init(&output.socket);
	output.socket.fd = STDOUT_FILENO;
	output.socket.dont_read = 1;
	output.socket.on_write = &st_output_on_write;
	output.socket.on_close = &st_output_on_close;
	output.running = 1;

	if(io_socket_connect(&thesaurus.socket, INADDR_LOOPBACK, 49152)) {
		thesaurus.running = 0;
	} else {
		thesaurus.running = 1;

		/* Only add the thesaurus connection if it connects successfully. */
		io_loop_add(&loop, &thesaurus.socket);
	}

	io_loop_add(&loop, &input.socket);
	io_loop_add(&loop, &output.socket);

	st_cgi_init();

	while(thesaurus.running || input.running || output.running) {
#if 0
		fprintf(stderr, "%d %d %d\n",
			thesaurus.running, input.running, output.running);
#endif

		/* CGI requires special handling; the server is not required to send
		 * end-of-file, therefore we must heed the "Content-Length" header
		 * and terminate the connection ourselves when the correct number of
		 * bytes has been received. */
		if(input.running
			&& input.cgi
			&& input.bytes_received >= input.content_length) {
			input.socket.close = 1;
		}

		io_loop_select(&loop);
	}

	io_loop_remove(&loop, &input.socket);
	io_loop_remove(&loop, &output.socket);
	/* This does not hurt, even if the socket never was in the loop.
	 * The socket calls LIST_INIT() in its init() function, thus the
	 * LIST_REMOVE() is not harmful. */
	io_loop_remove(&loop, &thesaurus.socket);

	if(input.running)
		io_socket_deinit(&input.socket);

	if(output.running)
		io_socket_deinit(&output.socket);

	io_loop_deinit(&loop);

	return 0;
}
