use strict;
use warnings;

use DBI;
use Text::Iconv;

# For character set conversion
my $iconv = Text::Iconv->new('ISO-8859-1', 'UTF-8');

my $nel = DBI->connect('dbi:mysql:nel', 'root', 'bjarne');
my %nel_terms;
my %nel_synonyms;
nel_load_terms();
nel_load_words();
$nel->disconnect;

my $masterwww = DBI->connect('dbi:mysql:masterwww', 'root', 'bjarne');
my %masterwww_terms;
my %masterwww_synonyms;
masterwww_load_words();
masterwww_load_stopwords();
$masterwww->disconnect;

my $minjournal2 = DBI->connect('dbi:mysql:minjournal2', 'root', 'bjarne');
minjournal2_save(1, "NEL", \%nel_terms, \%nel_synonyms);
minjournal2_save(2, "masterwww", \%masterwww_terms, \%masterwww_synonyms);
$minjournal2->disconnect;

# Load terms that have definitions
sub masterwww_load_words {
	my $stmt = 'SELECT word.id AS id, word AS word, definition AS explanation '
		. 'FROM word '
		. 'INNER JOIN definition ON definition_id = definition.id';
	my $q = $masterwww->prepare($stmt);
	$q->execute();

	while(my($id, $word, $explanation) = $q->fetchrow_array) {
		$masterwww_terms{'word_' . $id} = [$iconv->convert($word), $iconv->convert($explanation), undef];
	}
}

# Load terms that don't have definitions (stopwords)
sub masterwww_load_stopwords {
	my $stmt;
	my $q;

	$stmt = 'SELECT id, word FROM stopword';
	$q = $masterwww->prepare($stmt);
	$q->execute();

	while(my($id, $word) = $q->fetchrow_array) {
		$masterwww_terms{'stopword_' . $id} = [$iconv->convert($word), undef, undef];
	}

	$stmt = 'SELECT word.id AS id, word AS word '
		. 'FROM word '
		. 'WHERE definition_id IS NULL';
	$q = $masterwww->prepare($stmt);
	$q->execute();

	while(my($id, $word) = $q->fetchrow_array) {
		$masterwww_terms{'word_' . $id} = [$iconv->convert($word), undef, undef];
	}

	$stmt = 'SELECT id, misspell AS word, word_id AS term FROM misspell';
	$q = $masterwww->prepare($stmt);
	$q->execute();

	while(my($id, $word, $term) = $q->fetchrow_array) {
		$masterwww_synonyms{$id} = [$iconv->convert($word), 'word_' . $term];
	}
}

# Terms
sub nel_load_terms {
	my $stmt = 'SELECT id, term AS word, description AS explanation '
		. 'FROM terms';
	my $q = $nel->prepare($stmt);
	$q->execute();

	while(my($id, $word, $explanation) = $q->fetchrow_array) {
		# Already UTF-8
		$nel_terms{$id} = [$word, $explanation, undef];
	}
}

# Synonyms
sub nel_load_words {
	my $stmt = 'SELECT id, word, terms_id AS term '
		. 'FROM words';
	my $q = $nel->prepare($stmt);
	$q->execute();

	while(my($id, $word, $term) = $q->fetchrow_array) {
		$nel_synonyms{$id} = [$word, $term];
	}
}

sub minjournal2_save {
	my $thesaurus_id = shift;
	my $name = shift;
	my $terms = shift;
	my $synonyms = shift;

	printf "Saving thesaurus %s\n", $name;

	minjournal2_clear_thesaurus($thesaurus_id);
	minjournal2_create_thesaurus($thesaurus_id, $name);
	minjournal2_save_terms($thesaurus_id, $terms, $synonyms);
	minjournal2_save_synonyms($thesaurus_id, $terms, $synonyms);
}

sub minjournal2_clear_thesaurus {
	my $thesaurus_id = shift;

	my $stmt;
	my $q;

	$stmt = 'DELETE FROM Term WHERE thesaurus = ?';
	$q = $minjournal2->prepare($stmt);
	$q->execute($thesaurus_id);

	$stmt = 'DELETE FROM Synonym WHERE thesaurus = ?';
	$q = $minjournal2->prepare($stmt);
	$q->execute($thesaurus_id);

	$stmt = 'DELETE FROM Thesaurus WHERE id = ?';
	$q = $minjournal2->prepare($stmt);
	$q->execute($thesaurus_id);
}

sub minjournal2_create_thesaurus {
	my $thesaurus_id = shift;
	my $thesaurus_name = shift;

	my $stmt = 'INSERT INTO Thesaurus (id, name) '
		. 'VALUES (?, ?)';
	my $q = $minjournal2->prepare($stmt);
	$q->execute($thesaurus_id, $thesaurus_name);
}

sub minjournal2_save_terms {
	my $thesaurus_id = shift;
	my $terms = shift;
	my $synonyms = shift;

	for my $term_id (keys %$terms) {
		my $term = $terms->{$term_id};

		# Remove leading/trailing whitespace
		$term->[0] =~ s/^\s+//;
		$term->[0] =~ s/\s+$//;
		if(defined $term->[1]) {
			$term->[1] =~ s/^\s+//;
			$term->[1] =~ s/\s+$//;
		}

		my $stmt = 'INSERT INTO Term (thesaurus, word, explanation) '
			. 'VALUES (?, ?, ?)';
		my $q = $minjournal2->prepare($stmt);
		$q->execute($thesaurus_id, $term->[0], $term->[1]);

		$term->[2] = $minjournal2->last_insert_id(undef, undef, undef, undef);
	}
}

sub minjournal2_save_synonyms {
	my $thesaurus_id = shift;
	my $terms = shift;
	my $synonyms = shift;

	for my $synonym_id (keys %$synonyms) {
		my $synonym = $synonyms->{$synonym_id};
		my $term_id = $terms->{$synonym->[1]}->[2];

		if(!defined $term_id) {
			printf "warning: No definition for: %s\n", $synonym->[0];
			next;
		}

		$synonym->[0] =~ s/^\s+//;
		$synonym->[0] =~ s/\s+$//;

		my $stmt = 'INSERT INTO Synonym (thesaurus, term, word) '
			. 'VALUES (?, ?, ?)';
		my $q = $minjournal2->prepare($stmt);
		$q->execute($thesaurus_id, $term_id, $synonym->[0]);
	}
}
