#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "libs/buffer.h"
#include "libs/xmalloc.h"

void
buffer_init(struct buffer *buffer)
{
	assert(buffer);

	buffer->i = 0;
	buffer->size = 0;
	buffer->data = NULL;
}

/* Specify free_data = 1 to transfer ownership of the buffer data to the
 * caller. */
void
buffer_deinit(struct buffer *buffer, unsigned int free_data)
{
	assert(buffer);

	if(free_data)
		free(buffer->data);
}

/* The buffer is too small and must be resized. This function takes the
 * number of bytes that was attempted to add to the buffer, and computes
 * a new size for the buffer. This function must return AT LEAST
 * buffer->size + size. */
static unsigned int
st_compute_new_size(struct buffer *buffer, size_t size)
{
	assert(buffer);

	/* Always return AT LEAST size. */
	return size + buffer->size + buffer->size;
}

/* Append some data to the end of buffer. size is the number of bytes. */
void
buffer_write(struct buffer *buffer, const char *data, size_t size)
{
	assert(buffer);
	/* XXX: Assert on data if size > 0 */

	if(buffer->i + size > buffer->size) {
		/* Resize buffer */
		buffer->size = st_compute_new_size(buffer, size);
		buffer->data = xrealloc(buffer->data, buffer->size);
	}

	memcpy(buffer->data + buffer->i, data, size);
	buffer->i += size;
}

void
buffer_write_char(struct buffer *buffer, char ch)
{
	assert(buffer);

	buffer_write(buffer, &ch, 1);
} 

void
buffer_write_string(struct buffer *buffer, const char *string)
{
	assert(buffer);
	assert(string);

	buffer_write(buffer, string, strlen(string));
}

void
buffer_write_format(struct buffer *buffer, const char *format, ...)
{
	va_list ap;
	char *s;
	int l1, l2;

	va_start(ap, format);
	l1 = vsnprintf(NULL, 0, format, ap);
	va_end(ap);

	/* Negative means error. (man 3 vsnprintf) */
	assert(l1 >= 0);

	s = xmalloc(l1 + 1);

	va_start(ap, format);
	l2 = vsnprintf(s, l1 + 1, format, ap);
	va_end(ap);

	/* Same format, same arguments. If something goes wrong, it has to
	 * be a programming error. */
	assert(l1 == l2);

	buffer_write(buffer, s, l2);
	free(s);
}

void
buffer_move(struct buffer *buffer, char **out, unsigned int *out_length)
{
	*out = buffer->data;
	*out_length = buffer->i;

	/* Re-initialize the buffer so it doesn't refer to the memory we
	 * just moved. Also make sure that it is consistent. */
	buffer->i = 0;
	buffer->size = 0;
	buffer->data = NULL;
}

void
buffer_duplicate(struct buffer *buffer, char **out, unsigned int *out_length)
{
	char *dup;

	if(buffer->i == 0) {
		*out = NULL;
		*out_length = 0;
		return;
	}

	dup = xmalloc(buffer->i);
	memcpy(dup, buffer->data, buffer->i);

	*out = dup;
	*out_length = buffer->i;
}
