#ifndef WIN32
#include <sys/select.h>
#include <sys/time.h>
#endif
#include <sys/types.h>

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef WIN32
#include <stdint.h>
#include <ws2tcpip.h>
#endif

#include "libs/io-socket.h"
#include "libs/io-socket-buffer.h"
#include "libs/io-loop.h"
#include "libs/list.h"

void
io_loop_init(struct io_loop *io_loop)
{
	assert(io_loop);

	FD_ZERO(&io_loop->rfds);
	FD_ZERO(&io_loop->wfds);
	LIST_INIT(&io_loop->sockets);
}

void
io_loop_deinit(struct io_loop *io_loop)
{
	struct io_socket *socket;
	struct io_socket *socket_temp;

	assert(io_loop);

	/* Sockets that weren't allocated dynamically MUST be removed from
	 * the io loop prior to calling io_loop_deinit(). This is really
	 * no problem, however, as long as the programmer remembers to do
	 * do this. */
	LIST_FOR_SAFE(&io_loop->sockets, socket, socket_temp) {
		io_socket_close(socket);
		io_socket_deinit(socket);
		free(socket);
	}
}

void
io_loop_add(struct io_loop *io_loop, struct io_socket *socket)
{
	assert(io_loop);
	assert(socket);

	LIST_INSERT_BEFORE(&io_loop->sockets, socket);
	socket->io_loop = io_loop;
}

void
io_loop_remove(struct io_loop *io_loop, struct io_socket *socket)
{
	assert(io_loop);
	assert(socket);

	LIST_REMOVE(socket);
	socket->io_loop = NULL;
}

void
io_loop_select(struct io_loop *io_loop)
{
	struct io_socket *socket;
	struct io_socket *socket_temp;
	int maxfd;

	assert(io_loop);

	maxfd = -1;

	/* It's probably safer to zero them out at the beginning, although it
	 * should be possible just to clear the sockets we've marked. In that
	 * case, care should be taken to ensure that descriptors are cleared
	 * from both sets when they are removed. */
	FD_ZERO(&io_loop->rfds);
	FD_ZERO(&io_loop->wfds);

	LIST_FOR_SAFE(&io_loop->sockets, socket, socket_temp) {
		if(socket->close
			&& io_socket_buffer_is_empty(&socket->ibuffer)
			&& io_socket_buffer_is_empty(&socket->obuffer)) {
			io_socket_close(socket);
			/* We have to return, since the close callback may
			 * close additional sockets. */
			return;
		}
	}

	/* Find the sockets that we want to watch */
	LIST_FOR_SAFE(&io_loop->sockets, socket, socket_temp) {
		switch(socket->type) {
		case CLIENT:
			if(socket->fd > maxfd)
				maxfd = socket->fd;

			if(!io_socket_buffer_is_full(&socket->ibuffer)
				&& !socket->dont_read)
				FD_SET(socket->fd, &io_loop->rfds);

			if(!io_socket_buffer_is_empty(&socket->obuffer)
				&& !socket->dont_write)
				FD_SET(socket->fd, &io_loop->wfds);

			break;
		case SERVER:
			if(socket->fd > maxfd)
				maxfd = socket->fd;

			FD_SET(socket->fd, &io_loop->rfds);
			break;
		}
	}

	/* No sockets to select on, return lest we're in for an endless
	 * loop! */
	if(maxfd == -1)
		return;

	if(select(maxfd + 1, &io_loop->rfds, &io_loop->wfds, NULL, NULL) == -1) {
		if(errno == EINTR)
			return;

		fprintf(stderr, "warning: select() failed: %s", strerror(errno));
		return;
	}

	/* Readable sockets */
	LIST_FOR_SAFE(&io_loop->sockets, socket, socket_temp) {
		if(!FD_ISSET(socket->fd, &io_loop->rfds))
			continue;

		FD_CLR(socket->fd, &io_loop->rfds);

		switch(socket->type) {
		case CLIENT:
			io_socket_read(socket);
			break;
		case SERVER:
			io_socket_accept(socket);
			break;
		}
	}

	/* Writable sockets */
	LIST_FOR_SAFE(&io_loop->sockets, socket, socket_temp) {
		if(!FD_ISSET(socket->fd, &io_loop->wfds))
			continue;

		FD_CLR(socket->fd, &io_loop->wfds);

		switch(socket->type) {
		case CLIENT:
			io_socket_write(socket);
			break;
		case SERVER:
			break;
		}
	}
}
