#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "libs/buffer.h"
#include "libs/io-socket-buffer.h"
#include "libs/list.h"
#include "libs/xmalloc.h"

static void
st_io_socket_buffer_element_init(struct io_socket_buffer_element *e)
{
	assert(e);

	e->size = 0;
	e->data = NULL;

	e->rptr = 0;
}

static void
st_io_socket_buffer_element_deinit(struct io_socket_buffer_element *e)
{
	assert(e);

	free(e->data);
}

void
io_socket_buffer_init(struct io_socket_buffer *sb)
{
	assert(sb);

	LIST_INIT(&sb->elements);
}

void
io_socket_buffer_deinit(struct io_socket_buffer *sb)
{
	assert(sb);

	struct io_socket_buffer_element *e;
	struct io_socket_buffer_element *e_temp;

	LIST_FOR_SAFE(&sb->elements, e, e_temp) {
		st_io_socket_buffer_element_deinit(e);
		free(e);
	}
}

int
io_socket_buffer_is_full(struct io_socket_buffer *sb)
{
	assert(sb);

	return 0;
}

unsigned long int
io_socket_buffer_can_write(struct io_socket_buffer *sb)
{
	assert(sb);

	/* Arbitrary value. XXX: Maybe we could get this from getpagesize(),
	 * but we don't want to call that for every socket write. We don't
	 * want to just use an arbitrary value either. Perhaps called should
	 * specify a fallback, or perhaps we should just return a boolean. */
	return 4096;
}

/* This function doesn't duplicate the buffer, but uses it directly. The
 * caller should be prepared to give up its responsibility for the pointer.
 */
void
io_socket_buffer_write(struct io_socket_buffer *sb,
	unsigned char *data, unsigned long size)
{
	struct io_socket_buffer_element *e;

	assert(sb);

	if(!data || size == 0)
		return;

	e = xmalloc(sizeof(struct io_socket_buffer_element));
	st_io_socket_buffer_element_init(e);
	e->data = data;
	e->size = size;

	LIST_INSERT_BEFORE(&sb->elements, e);
}

/* Duplicate the buffer. The caller is still responsible for the original
 * data. */
void
io_socket_buffer_write_dup(struct io_socket_buffer *sb,
	const unsigned char *data, unsigned long size)
{
	struct io_socket_buffer_element *e;

	assert(sb);
	assert(data);

	if(size == 0)
		return;

	e = xmalloc(sizeof(struct io_socket_buffer_element));
	st_io_socket_buffer_element_init(e);
	e->size = size;
	e->data = xmalloc(size);
	memcpy(e->data, data, size);

	LIST_INSERT_BEFORE(&sb->elements, e);
}

void
io_socket_buffer_write_buffer(struct io_socket_buffer *sb,
	struct buffer *buffer)
{
	assert(sb);
	assert(buffer);

	if(buffer->i == 0)
		return;

	/* Take care of buffer's memory, reinitialize buffer. */
	io_socket_buffer_write(sb, (unsigned char *) buffer->data, buffer->i);
	buffer_deinit(buffer, 0);
	buffer_init(buffer);
}

int
io_socket_buffer_is_empty(struct io_socket_buffer *sb)
{
	return LIST_IS_EMPTY(&sb->elements);
}

unsigned long int
io_socket_buffer_can_read(struct io_socket_buffer *sb)
{
	struct io_socket_buffer_element *e;

	assert(sb);

	if(LIST_IS_EMPTY(&sb->elements))
		return 0;

	e = LIST_FIRST(&sb->elements);
	return e->size - e->rptr;
}

void
io_socket_buffer_read(struct io_socket_buffer *sb, unsigned char **data)
{
	struct io_socket_buffer_element *e;

	assert(sb);
	assert(data);

	if(LIST_IS_EMPTY(&sb->elements))
		return;

	e = LIST_FIRST(&sb->elements);
	*data = e->data + e->rptr;
}

void
io_socket_buffer_advance(struct io_socket_buffer *sb, unsigned long int size)
{
	struct io_socket_buffer_element *e;

	assert(sb);

	if(LIST_IS_EMPTY(&sb->elements))
		return;

	e = LIST_FIRST(&sb->elements);

	e->rptr += size;
	if(e->rptr == e->size) {
		LIST_REMOVE(e);
		st_io_socket_buffer_element_deinit(e);
		free(e);
	}
}

void
io_socket_buffer_read_dup(struct io_socket_buffer *sb, const unsigned char *data, unsigned long int size)
{
	struct io_socket_buffer_element *e;

	assert(sb);
	assert(data);

	if(size == 0)
		return;

	if(LIST_IS_EMPTY(&sb->elements))
		return;

	e = LIST_FIRST(&sb->elements);
}
