#ifndef WIN32
#include <netinet/in.h>
#include <netinet/ip.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#endif

#include <sys/types.h>

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef WIN32
#include <ws2tcpip.h>
#endif

#include "libs/io-loop.h"
#include "libs/io-socket.h"
#include "libs/io-socket-buffer.h"
#include "libs/list.h"
#include "libs/xmalloc.h"

void
io_socket_init(struct io_socket *socket)
{
	assert(socket);

	LIST_INIT(socket);
	socket->io_loop = NULL;
	socket->fd = -1;
	socket->addrlen = sizeof(&socket->addr);
	socket->type = CLIENT;
	socket->dont_read = 0;
	socket->dont_write = 0;
	socket->close = 0;
	io_socket_buffer_init(&socket->ibuffer);
	io_socket_buffer_init(&socket->obuffer);
	socket->data = NULL;
	socket->on_accept = NULL;
	socket->on_read = NULL;
	socket->on_write = NULL;
	socket->on_close = NULL;
}

void
io_socket_deinit(struct io_socket *socket)
{
	assert(socket);

	LIST_REMOVE(socket);

	io_socket_buffer_deinit(&socket->ibuffer);
	io_socket_buffer_deinit(&socket->obuffer);
}

void
io_socket_read(struct io_socket *socket)
{
	char *buffer;
	int can_read_socket;
	int can_write_buffer;
	int can_read;
	int len;

	assert(socket);

	can_read_socket = 1024;
	can_write_buffer = io_socket_buffer_can_write(&socket->ibuffer);

	can_read = can_read_socket < can_write_buffer ? can_read_socket : can_write_buffer;
	if(!can_read)
		return;

	buffer = xmalloc(can_read);
	len = read(socket->fd, buffer, can_read);
	if(len == -1) {
		fprintf(stderr, "error: read() failed: %s", strerror(errno));
		io_socket_close(socket);
		free(buffer);
		return;
	}

	if(len == 0) {
		io_socket_close(socket);
		free(buffer);
		return;
	}

	/* This can never fail since we're not writing any more than we CAN
	 * (write is limited to rbuffer_can_write()). */
	io_socket_buffer_write(&socket->ibuffer, (unsigned char *) buffer, len);

	if(socket->on_read)
		socket->on_read(socket);

	/* Don't free buffer. That is the resposibility of the io_socket_buffer. */
}

void
io_socket_write(struct io_socket *socket)
{
	unsigned char *buffer;
	int can_write_socket;
	int can_read_buffer;
	int can_write;
	int len;

	assert(socket);

	can_write_socket = 1024;
	can_read_buffer = io_socket_buffer_can_read(&socket->obuffer);

	can_write = can_write_socket < can_read_buffer ? can_write_socket : can_read_buffer;
	if(!can_write)
		return;

	io_socket_buffer_read(&socket->obuffer, &buffer);

	len = write(socket->fd, buffer, can_write);
	if(len == -1) {
		fprintf(stderr, "error: write() failed: %s\n", strerror(errno));
		io_socket_close(socket);
		return;
	}
	
	if(len == 0) {
		io_socket_close(socket);
		return;
	}

	if(socket->on_write)
		socket->on_write(socket);

	io_socket_buffer_advance(&socket->obuffer, len);
}

void
io_socket_close(struct io_socket *socket)
{
	assert(socket);
	assert(socket->fd != -1);

	while(close(socket->fd) == -1) {
		fprintf(stderr, "warning: close(): %s\n", strerror(errno));

		if(errno != EINTR)
			break;
	}

	if(socket->on_close)
		socket->on_close(socket);

	socket->fd = -1;
}

int
io_socket_open(struct io_socket *socket, const char *filename, int flags)
{
	int fd;

	assert(socket);

	fd = open(filename, flags);
	if(fd == -1) {
		fprintf(stderr, "error: open() failed: %s\n", strerror(errno));
		return 1;
	}

	socket->fd = fd;
	socket->type = CLIENT;
	return 0;
}

int
io_socket_connect(struct io_socket *sock, uint32_t host, unsigned short int port)
{
	int fd;
	struct sockaddr_in *addr;

	assert(sock);

	fd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if(fd == -1) {
		fprintf(stderr, "error: socket() failed: %s\n", strerror(errno));
		return 1;
	}

	sock->fd = fd;

	addr = (struct sockaddr_in *) &sock->addr;
	addr->sin_family = AF_INET;
	addr->sin_addr.s_addr = htonl(host);
	addr->sin_port = htons(port);

	sock->addrlen = sizeof(sock->addr);

	if(connect(fd, &sock->addr, sock->addrlen) == -1) {
		fprintf(stderr, "error: connect() failed: %s\n", strerror(errno));
		return 1;
	}

	sock->type = CLIENT;

	return 0;
}

int
io_socket_listen(struct io_socket *sock, unsigned short int port)
{
	int fd;
	struct sockaddr_in *addr;

	assert(sock);

	fd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if(fd == -1) {
		fprintf(stderr, "error: socket() failed: %s\n", strerror(errno));
		return 1;
	}

	sock->fd = fd;

	int reuseaddr = 1;
	if(setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &reuseaddr, sizeof(reuseaddr)) == -1) {
		fprintf(stderr, "warning: setsockopt() failed: %s\n", strerror(errno));
	}

	addr = (struct sockaddr_in *) &sock->addr;
	addr->sin_family = AF_INET;
	addr->sin_addr.s_addr = INADDR_ANY;
	addr->sin_port = htons(port);

	sock->addrlen = sizeof(sock->addr);

	sock->type = SERVER;

	if(bind(fd, &sock->addr, sock->addrlen) == -1) {
		fprintf(stderr, "error: bind() failed: %s\n", strerror(errno));
		return 1;
	}

	if(listen(fd, SOMAXCONN) == -1) {
		fprintf(stderr, "error: listen() failed: %s\n", strerror(errno));
		return 1;
	}

	return 0;
}

int
io_socket_accept(struct io_socket *socket)
{
	int fd;
	struct sockaddr addr;
	socklen_t addrlen;
	struct io_socket *client;

	assert(socket);

	addrlen = sizeof(addr);
	fd = accept(socket->fd, &addr, &addrlen);

	if(fd == -1) {
		fprintf(stderr, "error: accept() failed: %s\n", strerror(errno));
		return 1;
	}

	client = xmalloc(sizeof(struct io_socket));
	io_socket_init(client);

	client->fd = fd;
	memcpy(&client->addr, &addr, sizeof(addr));
	memcpy(&client->addrlen, &addrlen, sizeof(addrlen));
	client->type = CLIENT;

	io_loop_add(socket->io_loop, client);

	if(socket->on_accept)
		socket->on_accept(client);

	return 0;
}
