#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "libs/xmalloc.h"
#include "bloom-filter.h"

/* Modified Bob Jenkis hash*/
static unsigned long int
st_hash(const char *s, unsigned int l, unsigned int k)
{
	unsigned int i;
	unsigned long int p;

	p = 0x9e3779b9 + k;
	p += (p << 3);
	p ^= (p >> 11);
	p += (p << 15);

	for(i = 0; i < l; i++) {
		p += s[i];
		p += (p << 10);
		p ^= (p >> 6);
	}

	p += (p << 3);
	p ^= (p >> 11);
	p += (p << 15);

	return p;
}

/*
 * @param size bit array has 2^size bits
 */
void
bloom_filter_init(struct bloom_filter *filter, unsigned int size, unsigned int hashes)
{
	unsigned int i;
	unsigned int n;

	assert(filter);
	assert(size > 0);
	assert(hashes > 0);

	n = 1 << (size - 3);

	filter->size = size;
	filter->array = xcalloc(1, n);
	filter->hashes = hashes;

	for(i = 0; i < n; i++)
		filter->array[i] = 0;
}

void
bloom_filter_deinit(struct bloom_filter *filter)
{
	assert(filter);

	free(filter->array);
}

static void
st_set(unsigned char *a, unsigned long int p)
{
	unsigned long int p_byte, p_bit;

	p_byte = p >> 3;
	p_bit = p & 0x7;

	a[p_byte] |= (1 << p_bit);
}

static int
st_test(unsigned char *a, unsigned long int p)
{
	unsigned long int p_byte, p_bit;

	p_byte = p >> 3;
	p_bit = p & 0x7;

	return a[p_byte] & (1 << p_bit);
}

void
bloom_filter_insert(struct bloom_filter *bloom, const char *s, unsigned int l)
{
	unsigned int i;

	assert(bloom);
	assert(s);
	assert(l > 0);

	for(i = 0; i < bloom->hashes; i++)
		st_set(bloom->array, st_hash(s, l, i) & ((1 << bloom->size) - 1));
}

int
bloom_filter_contains(struct bloom_filter *bloom, const char *s, unsigned int l)
{
	unsigned int i;

	assert(bloom);
	assert(s);
	assert(l > 0);

	for(i = 0; i < bloom->hashes; i++) {
		if(!st_test(bloom->array, st_hash(s, l, i) & ((1 << bloom->size) - 1)))
			return 0;
	}

	return 1;
}
