#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include <mysql/mysql.h>

#include "libs/xmalloc.h"
#include "database.h"

void
database_init(struct database *db)
{
	assert(db);

	mysql_library_init(0, NULL, NULL);
}

void
database_deinit(struct database *db)
{
	assert(db);

	mysql_library_end();
}

void
database_connect(struct database *db, const char *options_group)
{
	assert(db);
	assert(options_group);

	db->mysql = mysql_init(NULL);
	if(!db->mysql) {
		fprintf(stderr, "error: Out of memory\n");
		exit(1);
	}

	if(mysql_options(db->mysql, MYSQL_READ_DEFAULT_GROUP, options_group)) {
		fprintf(stderr, "error: %s\n", mysql_error(db->mysql));
		exit(1);
	}

	if(!mysql_real_connect(db->mysql, NULL, NULL, NULL, NULL, 0, NULL, 0)) {
		fprintf(stderr, "error: %s\n", mysql_error(db->mysql));
		exit(1);
	}
}

void
database_disconnect(struct database *db)
{
	mysql_close(db->mysql);
}

static void
st_load_categories(struct database *db,
	void (*callback)(char *s[], unsigned long l[]))
{
	static const char query[] =
		"SELECT id, parent, name "
		"FROM Category";

	int err;
	MYSQL_RES *res;
	MYSQL_ROW row;
	unsigned long *len;

	err = mysql_real_query(db->mysql, query, sizeof(query));
	if(err) {
		fprintf(stderr, "error: %s\n", mysql_error(db->mysql));
		fprintf(stderr, "query was: %s\n", query);
		exit(1);
	}

	res = mysql_use_result(db->mysql);
	if(!res) {
		fprintf(stderr, "error: %s\n", mysql_error(db->mysql));
		exit(1);
	}

	while((row = mysql_fetch_row(res))) {
		len = mysql_fetch_lengths(res);

		callback(row, len);
	}

	mysql_free_result(res);
}

static void
st_load_thesauruses(struct database *db,
	void (*callback)(char *s[], unsigned long l[]))
{
	static const char query[] =
		"SELECT id, category, name "
		"FROM Thesaurus";

	int err;
	MYSQL_RES *res;
	MYSQL_ROW row;
	unsigned long *len;

	err = mysql_real_query(db->mysql, query, sizeof(query));
	if(err) {
		fprintf(stderr, "error: %s\n", mysql_error(db->mysql));
		fprintf(stderr, "query was: %s\n", query);
		exit(1);
	}

	res = mysql_use_result(db->mysql);
	if(!res) {
		fprintf(stderr, "error: %s\n", mysql_error(db->mysql));
		exit(1);
	}

	while((row = mysql_fetch_row(res))) {
		len = mysql_fetch_lengths(res);

		callback(row, len);
	}

	mysql_free_result(res);
}

static void
st_load_terms(struct database *db,
	void (*callback)(char *s[], unsigned long l[]))
{
	/* If declared as "const char *query", sizeof(query) will return the
	 * pointer size instead of the actual string length. */
	static const char query[] =
		"SELECT id, thesaurus, word, explanation "
		"FROM Term";

	int err;
	MYSQL_RES *res;
	MYSQL_ROW row;
	unsigned long *len;

	err = mysql_real_query(db->mysql, query, sizeof(query));
	if(err) {
		fprintf(stderr, "error: %s\n", mysql_error(db->mysql));
		fprintf(stderr, "query was: %s\n", query);
		exit(1);
	}

	res = mysql_use_result(db->mysql);
	if(!res) {
		fprintf(stderr, "error: %s\n", mysql_error(db->mysql));
		exit(1);
	}

	while((row = mysql_fetch_row(res))) {
		len = mysql_fetch_lengths(res);

		callback(row, len);
	}

	mysql_free_result(res);
}

static void
st_load_synonyms(struct database *db,
	void (*callback)(char *s[], unsigned long l[]))
{
	static const char query[] =
		"SELECT id, thesaurus, term, word "
		"FROM Synonym";

	int err;
	MYSQL_RES *res;
	MYSQL_ROW row;
	unsigned long *len;

	err = mysql_real_query(db->mysql, query, sizeof(query));
	if(err) {
		fprintf(stderr, "error: %s\n", mysql_error(db->mysql));
		fprintf(stderr, "query was: %s\n", query);
		exit(1);
	}

	res = mysql_use_result(db->mysql);
	if(!res) {
		fprintf(stderr, "error: %s\n", mysql_error(db->mysql));
		exit(1);
	}

	while((row = mysql_fetch_row(res))) {
		len = mysql_fetch_lengths(res);

		callback(row, len);
	}

	mysql_free_result(res);
}

void
database_load(struct database *db,
	void (*category_callback)(char *s[], unsigned long l[]),
	void (*thesaurus_callback)(char *s[], unsigned long l[]),
	void (*term_callback)(char *s[], unsigned long l[]),
	void (*synonym_callback)(char *s[], unsigned long l[]))
{
	st_load_categories(db, category_callback);
	st_load_thesauruses(db, thesaurus_callback);
	st_load_terms(db, term_callback);
	st_load_synonyms(db, synonym_callback);
}
