#ifndef WIN32
#include <sys/time.h>
#include <signal.h>
#include <time.h>
#endif

#include <assert.h>
#include <getopt.h>
#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <config.h>

#include <libxml/parser.h>

#include "libs/buffer.h"
#include "libs/io-loop.h"
#include "libs/io-socket.h"
#include "libs/io-socket-buffer.h"
#include "libs/list.h"
#include "libs/xmalloc.h"

#include "bloom-filter.h"
#include "category.h"
#include "category-hashtable.h"
#include "connection.h"
#include "database.h"
#include "levenshtein.h"
#include "normalize-no.h"
#include "result.h"
#include "search.h"
#include "stem-no.h"
#include "synonym-array.h"
#include "synonym.h"
#include "term.h"
#include "term-hashtable.h"
#include "thesaurus.h"
#include "thesaurus-hashtable.h"
#include "translate.h"
#include "translate-la-no.h"
#include "trie.h"

static struct database database;

struct category_hashtable categories;
struct thesaurus_hashtable thesauruses;
struct term_hashtable terms;

static char *
st_dup(const char *s, unsigned long int l)
{
	char *r;

	r = xmalloc(l);
	memcpy(r, s, l);
	return r;
}

static struct category *
st_get_category(unsigned long int id)
{
	struct category *cat;

	cat = category_hashtable_get(&categories, id);
	if(cat)
		return cat;

	cat = xmalloc(sizeof(*cat));
	category_init(cat);
	cat->id = id;

	category_hashtable_put(&categories, cat);
	return cat;
}

/* Category: id, parent, name */
static void
st_load_category(char *s[], unsigned long l[])
{
	struct category *category;

	category = st_get_category(atoi(s[0]));
	category->parent = st_get_category(atoi(s[1]));
	category->name = st_dup(s[3], l[3]);
}

static struct thesaurus *
st_get_thesaurus(unsigned long int id)
{
	struct thesaurus *thes;

	thes = thesaurus_hashtable_get(&thesauruses, id);
	if(thes)
		return thes;

	thes = xmalloc(sizeof(*thes));
	thesaurus_init(thes);
	thes->id = id;

	thesaurus_hashtable_put(&thesauruses, thes);
	return thes;
}

/* Thesaurus: id, category, name */
static void
st_load_thesaurus(char *s[], unsigned long l[])
{
	struct thesaurus *thesaurus;

	thesaurus = st_get_thesaurus(atoi(s[0]));
	thesaurus->category = st_get_category(atoi(s[1]));
	thesaurus->name = st_dup(s[2], l[2]);
	thesaurus->name_length = l[2];
}

static struct term *
st_get_term(unsigned long int id)
{
	struct term *term;

	term = term_hashtable_get(&terms, id);
	if(term)
		return term;

	term = xmalloc(sizeof(*term));
	term_init(term);
	term->id = id;

	term_hashtable_put(&terms, term);
	return term;
}

/* XXX: Logic should be moved out, similarly to search.c */
static void
st_insert_synonym(const struct synonym *synonym)
{
	struct synonym *s_normalized;
	struct synonym *s_translated;
	struct synonym *s_stemmed;

	char *normalized;
	unsigned int normalized_length;

	char *translated;
	unsigned int translated_length;

	char *stemmed;
	unsigned int stemmed_length;

	assert(synonym);

	/* XXX: Hard-coded norwegian language. */
	normalize_no(synonym->word, synonym->word_length,
		&normalized, &normalized_length);

	translate_la_no(
		normalized, normalized_length,
		&translated, &translated_length);

	stem_no(normalized, normalized_length,
		&stemmed, &stemmed_length);

	if(normalized_length) {
		s_normalized = xmalloc(sizeof(struct synonym));
		synonym_init(s_normalized);
		s_normalized->id = synonym->id;
		s_normalized->thesaurus = synonym->thesaurus;
		s_normalized->term = synonym->term;
		s_normalized->word = normalized;
		s_normalized->word_length = normalized_length;
		thesaurus_insert_synonym(synonym->thesaurus, s_normalized);
	}

	if(translated_length) {
		s_translated = xmalloc(sizeof(struct synonym));
		synonym_init(s_translated);
		s_translated->id = synonym->id;
		s_translated->thesaurus = synonym->thesaurus;
		s_translated->term = synonym->term;
		s_translated->word = translated;
		s_translated->word_length = translated_length;
		s_translated->transformations.translated = 1;
		thesaurus_insert_synonym(synonym->thesaurus, s_translated);
	}

	if(stemmed_length) {
		s_stemmed = xmalloc(sizeof(struct synonym));
		synonym_init(s_stemmed);
		s_stemmed->id = synonym->id;
		s_stemmed->thesaurus = synonym->thesaurus;
		s_stemmed->term = synonym->term;
		s_stemmed->word = stemmed;
		s_stemmed->word_length = stemmed_length;
		s_stemmed->transformations.stemmed = 1;
		thesaurus_insert_synonym(synonym->thesaurus, s_stemmed);
	}
}

/* Term: id, thesaurus, word, explanation */
static void
st_load_term(char *s[], unsigned long l[])
{
	struct term *term;
	struct synonym synonym;

	term = st_get_term(atoi(s[0]));
	term->thesaurus = st_get_thesaurus(atoi(s[1]));
	term->word = st_dup(s[2], l[2]);
	term->word_length = l[2];

	if(s[3]) {	
		term->explanation = st_dup(s[3], l[3]);
		term->explanation_length = l[3];
	} else {
		term->explanation = NULL;
		term->explanation_length = 0;
	}

	synonym_init(&synonym);
	synonym.id = 0; /* Generated ID */
	synonym.thesaurus = term->thesaurus;
	synonym.term = term;
	synonym.word = term->word;
	synonym.word_length = term->word_length;

	st_insert_synonym(&synonym);
}

/* Synonym: id, thesaurus, term, word */
static void
st_load_synonym(char *s[], unsigned long l[])
{
	struct synonym synonym;

	/* Note: Don't call st_dup() for the strings below. We're not going
	 * to free them, since they're simply used for copying into the
	 * actual structures. We don't even add this synonym to any index. */
	synonym_init(&synonym);
	synonym.id = atoi(s[0]);
	synonym.thesaurus = st_get_thesaurus(atoi(s[1]));
	synonym.term = st_get_term(atoi(s[2]));
	synonym.word = s[3];
	synonym.word_length = l[3];

	st_insert_synonym(&synonym);
}

/* Prepare data structures for searching. */
static void
st_prepare_index()
{
	unsigned int i;
	struct thesaurus *bucket;
	struct thesaurus *thesaurus;

	/* XXX: Get rid of the hard-coded bucket count. This should be stored
	 * as a member in struct thesaurus_hashtable, and perhaps be
	 * configurable at init time. */
	for(i = 0; i < 256; i++) {
		bucket = &thesauruses.buckets[i];

		LIST_FOR(bucket, thesaurus)
			trie_sort(&thesaurus->trie);
	}
}

static void
st_socket_on_read(struct io_socket *socket)
{
	struct connection *connection;
	unsigned long int can_read;
	unsigned char *buf;

	assert(socket);
	connection = socket->data;

	can_read = io_socket_buffer_can_read(&socket->ibuffer);
	io_socket_buffer_read(&socket->ibuffer, &buf);
	connection_handle(connection, buf, can_read);
	io_socket_buffer_advance(&socket->ibuffer, can_read);

	if(connection->request_parser.state == REQUEST_PARSER_ERROR) {
		printf("parse error: %s\n",
			connection->request_parser.error);
		socket->close = 1;
		return;
	}

	/* Only disconnect if the buffer has been flushed. */
	if(connection->request_parser.state == REQUEST_PARSER_DONE) {
		printf("done\n");
		socket->close = 1;
		return;
	}
}

static void
st_socket_on_write(struct io_socket *socket)
{
	struct connection *connection;

	assert(socket);
	connection = socket->data;
}

static void
st_socket_on_close(struct io_socket *socket)
{
	struct connection *connection;

	assert(socket);
	connection = socket->data;

	/* XXX: Show IP information, log event. */
	fprintf(stderr, "disconnected\n");
	/*
	printf("client had %d lookups, %d results\n",
		connection->lookups,
		connection->results);
	*/

	io_socket_deinit(socket);

	connection_deinit(connection);
	free(connection);
}

static void
st_socket_on_accept(struct io_socket *socket)
{
	struct connection *connection;

	/* XXX: Show IP information, log event. */
	fprintf(stderr, "connected\n");

	connection = xmalloc(sizeof(struct connection));
	connection_init(connection);
	connection->socket = socket;

	socket->data = connection;
	socket->on_read = &st_socket_on_read;
	socket->on_write = &st_socket_on_write;
	socket->on_close = &st_socket_on_close;
}

static int running;

static void
st_stop(int n)
{
	fprintf(stderr, "Received signal %d\n", n);
	running = 0;
}

static void
st_init()
{
	/* Initialize internal structures */
	category_hashtable_init(&categories);
	thesaurus_hashtable_init(&thesauruses);
	term_hashtable_init(&terms);
}

static void
st_deinit()
{
	category_hashtable_deinit(&categories);
	thesaurus_hashtable_deinit(&thesauruses);
	term_hashtable_deinit(&terms);
}

int
main(int argc, char *argv[])
{
	static char *opt_database = "minjournal";

	struct io_socket socket;
	struct io_loop loop;

	/* Check that the libxml version in use is compatible with the
	 * version the software has been compiled against. */
	LIBXML_TEST_VERSION

	/* This is supposed to make the program portable. It looks like
	 * this call makes certain that the string manipulation functions
	 * obey the rules according to environment-variable locale
	 * settings. So then we just have to make sure the host
	 * environment is using UTF-8...
	 */
	setlocale(LC_ALL, "");

	/* Parse command-line arguments */
	while(1) {
		int c;

		c = getopt(argc, argv, "hVd:");
		if(c == -1)
			break;

		switch(c) {
		case 'h':
			printf("Usage: %s [OPTION]...\n", argv[0]);
			printf("\n");
			printf("  -h            display this help and exit\n");
			printf("  -V            output version information and exit\n");
			printf("\n");
			printf("  -d database   specify the database option group to use\n");
			printf("\n");
			exit(0);
			break;
		case 'V':
			printf("%s (%s) %s\n", PACKAGE_NAME, "minjournal", PACKAGE_VERSION);
			printf("Written by Vegard Nossum.\n");
			printf("\n");
			exit(0);
			break;
		case 'd':
			opt_database = optarg;
			break;
		}
	}

	if(optind < argc) {
		fprintf(stderr, "%s: invalid option.\n", argv[0]);
		fprintf(stderr, "Try `%s -h' for more information.\n", argv[0]);
		exit(1);
	}

	st_init();

	/* Read databases */
	database_init(&database);

	fprintf(stderr, "Connecting to the database... ");
	fflush(stderr);
	database_connect(&database, opt_database);
	fprintf(stderr, "ok\n");

	fprintf(stderr, "Loading data... ");
	fflush(stderr);
	database_load(&database,
		&st_load_category,
		&st_load_thesaurus,
		&st_load_term,
		&st_load_synonym);

	fprintf(stderr, "ok\n");
	database_disconnect(&database);
	database_deinit(&database);

	fprintf(stderr, "Sorting index... ");
	fflush(stderr);
	st_prepare_index();
	fprintf(stderr, "ok\n");

	fprintf(stderr, "Ready.\n");
	fprintf(stderr, "\n");

	/* Wait for requests, perform look-ups */
	io_socket_init(&socket);
	socket.on_accept = &st_socket_on_accept;

	if(io_socket_listen(&socket, 49152)) {
		/* XXX: Error message is already printed. Move to here? */
		exit(1);
	}

	io_loop_init(&loop);
	io_loop_add(&loop, &socket);

	running = 1;
#ifndef WIN32
	signal(SIGHUP, &st_stop);
	signal(SIGINT, &st_stop);
	signal(SIGTERM, &st_stop);
#endif

	/* Main loop */
	while(running)
		io_loop_select(&loop);

	io_loop_remove(&loop, &socket);
	io_socket_deinit(&socket);
	io_loop_deinit(&loop);

	st_deinit();
	return 0;
}
