#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libxml/parser.h>

#include "libs/buffer.h"
#include "libs/list.h"
#include "libs/xmalloc.h"
#include "libs/xml.h"

#include "request-parser.h"

/* XML request parser; see request.xsd */

void
request_parser_init(struct request_parser *parser)
{
	parser->state = REQUEST_PARSER_ROOT;
	parser->error = NULL;
}

void
request_parser_deinit(struct request_parser *parser)
{
	free(parser->error);
}

static void
st_startElement(void *data, const xmlChar *name, const xmlChar **attrs)
{
	struct request_parser *parser;
	const struct request_parser_handler *handler;

	assert(data);

	parser = data;
	handler = parser->handler;

	switch(parser->state) {
	case REQUEST_PARSER_ROOT:
		if(!strcmp((char *) name, "document")) {
			handler->on_start_document(parser->request);
			parser->state = REQUEST_PARSER_ROOT_DOCUMENT;
			break;
		}

		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_ROOT_DOCUMENT:
		if(!strcmp((char *) name, "sources")) {
			handler->on_start_sources(parser->request);
			parser->state = REQUEST_PARSER_ROOT_DOCUMENT_SOURCES;
			break;
		}

		if(!strcmp((char *) name, "text")) {
			handler->on_start_text(parser->request);
			parser->state = REQUEST_PARSER_ROOT_DOCUMENT_TEXT;
			break;
		}

		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_ROOT_DOCUMENT_SOURCES:
		if(!strcmp((char *) name, "category")) {
			handler->on_start_category(parser->request);
			parser->state = REQUEST_PARSER_ROOT_DOCUMENT_SOURCES_CATEGORY;
			break;
		}

		if(!strcmp((char *) name, "thesaurus")) {
			handler->on_start_thesaurus(parser->request);
			parser->state = REQUEST_PARSER_ROOT_DOCUMENT_SOURCES_THESAURUS;
			break;
		}

		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_ROOT_DOCUMENT_SOURCES_CATEGORY:
		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_ROOT_DOCUMENT_SOURCES_THESAURUS:
		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_ROOT_DOCUMENT_TEXT:
		if(!strcmp((char *) name, "word")) {
			handler->on_start_word(parser->request);
			parser->state = REQUEST_PARSER_ROOT_DOCUMENT_TEXT_WORD;
			break;
		}

		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_ROOT_DOCUMENT_TEXT_WORD:
		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_ERROR:
		return;
	case REQUEST_PARSER_DONE:
		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		return;
	}
}

static void
st_endElement(void *data, const xmlChar *name)
{
	struct request_parser *parser;
	const struct request_parser_handler *handler;

	assert(data);

	parser = data;
	handler = parser->handler;

	switch(parser->state) {
	case REQUEST_PARSER_ROOT:
		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_ROOT_DOCUMENT:
		if(!strcmp((char *) name, "document")) {
			handler->on_end_document(parser->request);
			parser->state = REQUEST_PARSER_DONE;
			break;
		}

		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_ROOT_DOCUMENT_SOURCES:
		if(!strcmp((char *) name, "sources")) {
			handler->on_end_sources(parser->request);
			parser->state = REQUEST_PARSER_ROOT_DOCUMENT;
			break;
		}

		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_ROOT_DOCUMENT_SOURCES_CATEGORY:
		if(!strcmp((char *) name, "category")) {
			handler->on_end_category(parser->request);
			parser->state = REQUEST_PARSER_ROOT_DOCUMENT_SOURCES;
			break;
		}

		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_ROOT_DOCUMENT_SOURCES_THESAURUS:
		if(!strcmp((char *) name, "thesaurus")) {
			handler->on_end_thesaurus(parser->request);
			parser->state = REQUEST_PARSER_ROOT_DOCUMENT_SOURCES;
			break;
		}

		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_ROOT_DOCUMENT_TEXT:
		if(!strcmp((char *) name, "text")) {
			handler->on_end_text(parser->request);
			parser->state = REQUEST_PARSER_ROOT_DOCUMENT;
			break;
		}

		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_ROOT_DOCUMENT_TEXT_WORD:
		if(!strcmp((char *) name, "word")) {
			handler->on_end_word(parser->request);
			parser->state = REQUEST_PARSER_ROOT_DOCUMENT_TEXT;
			break;
		}

		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_ERROR:
		return;
	case REQUEST_PARSER_DONE:
		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		return;
	}
}

static void
st_characters(void *data, const xmlChar *ch, int len)
{
	struct request_parser *parser;
	const struct request_parser_handler *handler;

	assert(data);

	parser = data;
	handler = parser->handler;

	switch(parser->state) {
	case REQUEST_PARSER_ROOT:
		/* ignore */
		break;
	case REQUEST_PARSER_ROOT_DOCUMENT:
		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_ROOT_DOCUMENT_SOURCES:
		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		break;
	case REQUEST_PARSER_ROOT_DOCUMENT_SOURCES_CATEGORY:
		handler->on_characters_category(parser->request,
			(const char *) ch, len);
		break;
	case REQUEST_PARSER_ROOT_DOCUMENT_SOURCES_THESAURUS:
		handler->on_characters_thesaurus(parser->request,
			(const char *) ch, len);
		break;
	case REQUEST_PARSER_ROOT_DOCUMENT_TEXT:
		handler->on_characters_text(parser->request,
			(const char *) ch, len);
		break;
	case REQUEST_PARSER_ROOT_DOCUMENT_TEXT_WORD:
		handler->on_characters_word(parser->request,
			(const char *) ch, len);
		break;
	case REQUEST_PARSER_ERROR:
		return;
	case REQUEST_PARSER_DONE:
		/* error */
		parser->state = REQUEST_PARSER_ERROR;
		return;
	}
}

static char *
st_vstrdup(const char *msg, va_list ap)
{
	char buf[1024];

	vsnprintf(buf, sizeof(buf), msg, ap);

	return strdup(buf);
}

static void
st_warning(void *data, const char *msg, ...)
{
	struct request_state *state;

	assert(data);

	state = data;
}

static void
st_error(void *data, const char *msg, ...)
{
	struct request_parser *parser;
	va_list ap;

	assert(data);

	parser = data;
	parser->state = REQUEST_PARSER_ERROR;

	va_start(ap, msg);
	parser->error = st_vstrdup(msg, ap);
	va_end(ap);
}

static void
st_fatalError(void *data, const char *msg, ...)
{
	struct request_parser *parser;
	va_list ap;

	assert(data);

	parser = data;
	parser->state = REQUEST_PARSER_ERROR;

	va_start(ap, msg);
	parser->error = st_vstrdup(msg, ap);
	va_end(ap);
}

const xmlSAXHandler request_parser_handler = {
	.startElement = &st_startElement,
	.endElement = &st_endElement,
	.characters = &st_characters,
	.warning = &st_warning,
	.error = &st_error,
	.fatalError = &st_fatalError,
};
