#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libxml/parser.h>

#include "libs/buffer.h"
#include "libs/list.h"
#include "libs/xmalloc.h"
#include "libs/xml.h"

#include "category.h"
#include "category-hashtable.h"
#include "request.h"
#include "request-parser.h"
#include "result.h"
#include "search.h"
#include "term.h"
#include "thesaurus.h"
#include "thesaurus-hashtable.h"

void
request_init(struct request *request)
{
	assert(request);

	LIST_INIT(&request->categories);
	LIST_INIT(&request->thesauruses);

	buffer_init(&request->output);
}

void
request_deinit(struct request *request)
{
	assert(request);

	buffer_deinit(&request->output, 1);
}

/* XXX: This is a bit ugly. They should not be global, but accessible through
 * a chain from the main program. */
extern struct category_hashtable categories;
extern struct thesaurus_hashtable thesauruses;
extern struct term_hashtable terms;

static unsigned long int
st_buf_to_uint(struct buffer *buffer)
{
	const char *s;
	unsigned int l;
	unsigned long int r;
	unsigned int i;

	s = buffer->data;
	l = buffer->i;
	r = 0;

	for(i = 0; i < l; i++) {
		if(s[i] >= '0' && s[i] <= '9')
			r = 10 * r + (s[i] - '0');
	}

	return r;
}

static void
st_add_category(struct request *request,
	unsigned long int category_id)
{
	struct category *category;
	struct request_source_category *source_category;

	assert(request);

	category = category_hashtable_get(&categories, category_id);
	if(!category)
		return;

	source_category = xmalloc(sizeof(struct request_source_category));
	source_category->category = category;

	LIST_INSERT_BEFORE(&request->categories, source_category);
}

static void
st_add_thesaurus(struct request *request,
	unsigned long int thesaurus_id)
{
	struct thesaurus *thesaurus;
	struct request_source_thesaurus *source_thesaurus;

	assert(request);

	thesaurus = thesaurus_hashtable_get(&thesauruses, thesaurus_id);
	if(!thesaurus)
		return;

	source_thesaurus = xmalloc(sizeof(struct request_source_thesaurus));
	source_thesaurus->thesaurus = thesaurus;

	LIST_INSERT_BEFORE(&request->thesauruses, source_thesaurus);
}


static void
st_handle_word(struct request *request, const char *s, unsigned int l)
{
	struct buffer *output;

	struct request_source_category *source_category;
	struct request_source_thesaurus *source_thesaurus;

	struct thesaurus **thesauruses;
	unsigned int thesaurus_n;
	unsigned int thesaurus_i;

	struct result result_list;
	struct result **result_array;
	unsigned int result_n;
	unsigned int result_i;

	assert(request);
	assert(s);

	output = &request->output;

	thesaurus_n = 0;
	LIST_FOR(&request->thesauruses, source_thesaurus)
		thesaurus_n++;

	thesauruses = xcalloc(thesaurus_n, sizeof(*thesauruses));

	thesaurus_i = 0;
	LIST_FOR(&request->thesauruses, source_thesaurus)
		thesauruses[thesaurus_i++] = source_thesaurus->thesaurus;

	search(thesauruses, thesaurus_n,
		&result_list,
		s, l);

	free(thesauruses);

	result_list_sort(&result_list, &result_array, &result_n);

	buffer_write_string(output, "<word>");

	buffer_write_string(output, "<original>");
	xml_escape_to_buffer(s, l, output);
	buffer_write_string(output, "</original>");

	for(result_i = 0; result_i < result_n; result_i++) {
		struct result *result;
		struct synonym *synonym;
		struct term *term;
		int exact;
		int stemmed;
		int translated;
		int distance;

		result = result_array[result_i];
		synonym = result->synonym;
		term = synonym->term;

		stemmed = result->transformations.stemmed;
		translated = result->transformations.translated;
		distance = result->distance;
		exact = (distance == 0) && !(stemmed || translated);

		buffer_write_format(output,
			"<result"
			" exact=\"%s\""
			" stemmed=\"%s\""
			" translated=\"%s\""
			" editDistance=\"%d\""
			">",
			exact ? "true" : "false",
			stemmed ? "true" : "false",
			translated ? "true" : "false",
			distance
		);

#if 0
		fprintf(stdout, "%.*s: %d %d %d %d\n", l, s, exact, stemmed, translated, distance);
#endif

		buffer_write_string(output, "<synonym>");
		xml_escape_to_buffer(synonym->word, synonym->word_length,
			output);
		buffer_write_string(output, "</synonym>");

		buffer_write_string(output, "<term>");
		xml_escape_to_buffer(term->word, term->word_length,
			output);
		buffer_write_string(output, "</term>");

		if(term->explanation) {
			buffer_write_string(output, "<explanation>");
			xml_escape_to_buffer(term->explanation,
				term->explanation_length,
				output);
			buffer_write_string(output, "</explanation>");
		}

		buffer_write_string(output, "<thesaurus>");
		buffer_write_format(output, "%d", term->thesaurus->id);
		buffer_write_string(output, "</thesaurus>");

		buffer_write_string(output, "</result>");

		free(result);
	}

	buffer_write_string(output, "</word>");

	free(result_array);
}

static void
on_start_document(struct request *request)
{
	buffer_write_string(&request->output, "<document>");
}

static void
on_end_document(struct request *request)
{
	buffer_write_string(&request->output, "</document>");
}

static void
on_start_sources(struct request *request)
{
}

static void
on_end_sources(struct request *request)
{
}

static void
on_start_category(struct request *request)
{
	/* begin category reading */
	buffer_init(&request->category_id);
}

static void
on_characters_category(struct request *request, const char *s, unsigned int l)
{
	buffer_write(&request->category_id, s, l);
}

static void
on_end_category(struct request *request)
{
	/* end category reading */
	st_add_category(request, st_buf_to_uint(&request->category_id));
	buffer_deinit(&request->category_id, 1);
}

static void
on_start_thesaurus(struct request *request)
{
	/* begin thesaurus reading */
	buffer_init(&request->thesaurus_id);
}

static void
on_characters_thesaurus(struct request *request, const char *s, unsigned int l)
{
	buffer_write(&request->thesaurus_id, s, l);
}

static void
on_end_thesaurus(struct request *request)
{
	/* end thesaurus reading */
	st_add_thesaurus(request, st_buf_to_uint(&request->thesaurus_id));
	buffer_deinit(&request->thesaurus_id, 1);
}

static void
on_start_text(struct request *request)
{
	/* begin text reading */
	buffer_write_string(&request->output, "<text>");
	buffer_init(&request->text);
}

static void
on_characters_text(struct request *request, const char *s, unsigned int l)
{
	buffer_write(&request->text, s, l);
}

static void
on_end_text(struct request *request)
{
	/* end text reading */
	xml_escape_buffer_to_buffer(&request->text, &request->output);
	buffer_deinit(&request->text, 1);
	buffer_write_string(&request->output, "</text>");
}

static void
on_start_word(struct request *request)
{
	/* end text reading */
	xml_escape_buffer_to_buffer(&request->text, &request->output);
	buffer_deinit(&request->text, 1);

	/* begin word reading */
	//buffer_write_string(&request->output, "<word>");
	buffer_init(&request->word);
}

static void
on_characters_word(struct request *request, const char *s, unsigned int l)
{
	buffer_write(&request->word, s, l);
}

static void
on_end_word(struct request *request)
{
	/* end word reading */
	st_handle_word(request,
		request->word.data, request->word.i);
	buffer_deinit(&request->word, 1);
	//buffer_write_string(&request->output, "</word>");

	/* begin text reading */
	buffer_init(&request->text);
}

const struct request_parser_handler request_handler = {
	.on_start_document		= &on_start_document,
	.on_start_sources		= &on_start_sources,
	.on_start_category		= &on_start_category,
	.on_start_thesaurus		= &on_start_thesaurus,
	.on_start_text			= &on_start_text,
	.on_start_word			= &on_start_word,
	.on_end_document		= &on_end_document,
	.on_end_sources			= &on_end_sources,
	.on_end_category		= &on_end_category,
	.on_end_thesaurus		= &on_end_thesaurus,
	.on_end_text			= &on_end_text,
	.on_end_word			= &on_end_word,
	.on_characters_category		= &on_characters_category,
	.on_characters_thesaurus	= &on_characters_thesaurus,
	.on_characters_text		= &on_characters_text,
	.on_characters_word		= &on_characters_word,
};
