#include <assert.h>
#include <stdlib.h>

#include "libs/list.h"
#include "libs/xmalloc.h"

#include "result.h"
#include "synonym.h"
#include "term.h"

void
result_init(struct result *result)
{
	assert(result);

	result->synonym = NULL;
	result->transformations.translated = 0;
	result->transformations.stemmed = 0;
	result->index = 0;
	result->index_a = 0;
	result->index_b = 0;
	result->index_ab = 0;
	result->distance = 0;
}

void
result_deinit(struct result *result)
{
	assert(result);
}

static unsigned int
st_result_score(const struct result *r)
{
	int score;

#if 0
	/* Slows down too much, even for debugging. */
	assert(r);
#endif

	/* Rationale:
	 *  add one if the word is translated
	 *  add one if the word is stemmed
	 *  add one for each edit (between the query and entry word)
	 *  add three if the full index was used
	 *  add two if the any of the single-letter indexes were used
	 *  add one if the two-letter index was used
	 */

	score = r->transformations.translated
		+ r->transformations.stemmed
		+ r->distance
		+ 3 * r->index
		+ 2 * (r->index_a + r->index_b)
		+ r->index_ab;

	return score;
}

/* This function sorts using the synonym id before the result score. This
 * way, we can keep only the best candidate for each synonym.
 */
static int
st_result_compare_unique(const void *_a, const void *_b)
{
	const struct result *a;
	const struct result *b;

	assert(_a);
	assert(_b);

	a = *((struct result *const *) _a);
	b = *((struct result *const *) _b);

	if(a->synonym->term->id < b->synonym->term->id)
		return -1;

	if(a->synonym->term->id > b->synonym->term->id)
		return 1;

	if(a->score < b->score)
		return -1;

	if(a->score > b->score)
		return 1;

	return 0;
}

static void
st_result_unique(struct result **array, unsigned int *result_n)
{
	int i;
	int j;
	int n;

	j = 0;
	n = *result_n;

	if(n == 0)
		return;

	for(i = 1; i < n; i++) {
		/* Delete the duplicates. We want no leaks! */
		if(array[i]->synonym->term->id == array[j]->synonym->term->id) {
			LIST_REMOVE(array[i]);
			result_deinit(array[i]);
			free(array[i]);
			continue;
		}

		array[++j] = array[i];
	}

	*result_n = j + 1;
}

static int
st_result_compare(const void *_a, const void *_b)
{
	const struct result *a;
	const struct result *b;

	assert(_a);
	assert(_b);

	a = *((struct result *const *) _a);
	b = *((struct result *const *) _b);

	if(a->score < b->score)
		return -1;

	if(a->score > b->score)
		return 1;

	return 0;
}

void
result_list_sort(struct result *result_list,
	struct result ***result_array, unsigned int *result_n)
{
	unsigned int i;
	unsigned int n;
	struct result *result;
	struct result **array;

	assert(result_list);
	assert(result_array);
	assert(result_n);

	n = 0;
	LIST_FOR(result_list, result)
		n++;

	if(n == 0) {
		/* Don't do anything else */
		*result_array = NULL;
		*result_n = 0;
		return;
	}

	array = xcalloc(n, sizeof(struct result *));

	i = 0;
	LIST_FOR(result_list, result) {
		result->score = st_result_score(result);
		array[i++] = result;
	}

	qsort(array, n, sizeof(struct result *), &st_result_compare_unique);
	st_result_unique(array, &n);
	qsort(array, n, sizeof(struct result *), &st_result_compare);

	*result_array = array;
	*result_n = n;
}
