#include <assert.h>
#include <stdlib.h>

#include "libs/list.h"

#include "normalize-no.h"
#include "result.h"
#include "stem-no.h"
#include "thesaurus.h"
#include "transformations.h"
#include "translate-la-no.h"

static void
st_search(struct thesaurus *thesauruses[], unsigned int thesauruses_n,
	struct result *results,
	char *const s[], unsigned int l[],
	struct transformations t[],
	unsigned int n)
{
	static void (*const algorithms[])(
		struct thesaurus *thesaurus, struct result *results,
		const char *s, unsigned int l,
		const struct transformations *t) = {
		&thesaurus_search_trie,
		&thesaurus_search_index_ab,
		/*
		&thesaurus_search_index_a,
		&thesaurus_search_index_b,
		*/
		NULL,
	};

	struct thesaurus *thesaurus;
	unsigned int k;
	unsigned int i;
	unsigned int j;

	/* Perform the exact-match search on all thesauruses first */
	for(k = 0; k < thesauruses_n; k++) {
		thesaurus = thesauruses[k];

		for(i = 0; i < n; i++) {
			if(l[i] == 0)
				continue;

			algorithms[0](thesaurus, results,
				s[i], l[i], &t[i]);
		}
	}

	/* Then perform fuzzy matching */
	for(k = 0; k < thesauruses_n; k++) {
		thesaurus = thesauruses[k];

		for(i = 0; i < n; i++) {
			if(l[i] == 0)
				continue;

			/* Return as soon as we have anything to report. */
			if(!LIST_IS_EMPTY(results))
				return;

			for(j = 1; algorithms[j]; j++) {
				algorithms[j](thesaurus, results,
					s[i], l[i], &t[i]);
			}
		}
	}
}

void
search(struct thesaurus *thesauruses[], unsigned int thesauruses_n,
	struct result *results, const char *s, unsigned int l)
{
	char *strings[3];
	unsigned int lengths[3];
	struct transformations transformations[3];

	LIST_INIT(results);

	if(!s || l == 0)
		return;

	/* Normalize the query string, i.e. capitalize letters and remove
	 * punctuation and non-letters. Use Norwegian table to preserve
	 * Norwegian extra letters */
	normalize_no(s, l, &strings[0], &lengths[0]);
	if(!strings[0] || lengths[0] == 0)
		return;

	transformations_init(&transformations[0], 0, 0);

	/* Translate the word using latin-to-norwegian transformation
	 * rules */
	translate_la_no(
		strings[0], lengths[0],
		&strings[1], &lengths[1]);
	transformations_init(&transformations[1], 1, 0);

	/* Stem the word using the norwegian stemmer */
	stem_no(strings[0], lengths[0],
		&strings[2], &lengths[2]);
	transformations_init(&transformations[2], 0, 1);

	st_search(thesauruses, thesauruses_n,
		results,
		strings, lengths, transformations, 3);

	free(strings[0]);
	free(strings[1]);
	free(strings[2]);
}
