#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "libs/list.h"
#include "libs/xmalloc.h"

#include "edit-distance.h"
#include "result.h"
#include "synonym.h"
#include "term.h"
#include "thesaurus.h"
#include "transformations.h"
#include "trie.h"

void
thesaurus_init(struct thesaurus *thesaurus)
{
	int i;

	assert(thesaurus);

	synonym_array_init(&thesaurus->index);

	for(i = 0; i < 26; i++) {
		synonym_array_init(&thesaurus->index_a[i]);
		synonym_array_init(&thesaurus->index_b[i]);
	}

	for(i = 0; i < 26 * 26; i++)
		synonym_array_init(&thesaurus->index_ab[i]);

	trie_init(&thesaurus->trie, 0);
}

void
thesaurus_deinit(struct thesaurus *thesaurus)
{
	unsigned int i;

	assert(thesaurus);

	free(thesaurus->name);

	for(i = 0; i < 26; i++) {
		synonym_array_deinit(&thesaurus->index_a[i], 0);
		synonym_array_deinit(&thesaurus->index_b[i], 0);
	}

	for(i = 0; i < 26 * 26; i++)
		synonym_array_deinit(&thesaurus->index_ab[i], 0);

	synonym_array_deinit(&thesaurus->index, 1);

	trie_deinit(&thesaurus->trie);
}

/* This call should be made unneccessary because strings should already
 * be normalized when they reach the insert() function.
 * 
 * Update: It's still necessary, since strings are 8-bit, but we have
 * 26-element arrays only.
 *
 * XXX: How to handle non-ASCII characters?
 */
static unsigned int
st_char_index(char c)
{
	if(c >= 'a' && c <= 'z')
		return c - 'a';

	if(c >= 'A' && c <= 'Z')
		return c - 'A';

	/* assert((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z')); */
	return 0;
}

void
thesaurus_insert_synonym(struct thesaurus *thesaurus, struct synonym *synonym)
{
	unsigned int i_a;
	unsigned int i_b;
	struct trie *trie;

	assert(thesaurus);
	assert(synonym);
	assert(synonym->word_length);

	synonym_array_insert(&thesaurus->index, synonym);

	i_a = st_char_index(synonym->word[0]);
	i_b = st_char_index(synonym->word[synonym->word_length - 1]);

	synonym_array_insert(&thesaurus->index_a[i_a], synonym);
	synonym_array_insert(&thesaurus->index_b[i_b], synonym);
	synonym_array_insert(&thesaurus->index_ab[i_a + 26 * i_b], synonym);

	trie = trie_insert(&thesaurus->trie, synonym->word, synonym->word_length);
	if(!trie->synonyms) {
		trie->synonyms = xmalloc(sizeof(struct synonym_array));
		synonym_array_init(trie->synonyms);
	}

	synonym_array_insert(trie->synonyms, synonym);
}

void
thesaurus_search_trie(struct thesaurus *thesaurus,
	struct result *result_list,
	const char *s, unsigned int l,
	const struct transformations *t)
{
	struct trie *trie;
	struct synonym_array *array;
	struct synonym *synonym;
	struct result *result;
	unsigned int i;

	trie = trie_search(&thesaurus->trie, s, l);
	if(trie && trie->synonyms) {
		array = trie->synonyms;

		for(i = 0; i < array->n; i++) {
			synonym = array->synonyms[i];

			result = xmalloc(sizeof(*result));
			result_init(result);
			result->synonym = synonym;
			transformations_or(&result->transformations,
				&synonym->transformations, t);

			LIST_INSERT_BEFORE(result_list, result);
		}
	}
}

void
thesaurus_search_index_a(struct thesaurus *thesaurus,
	struct result *result_list,
	const char *s, unsigned int l,
	const struct transformations *t)
{
	unsigned int i;
	unsigned int n;
	unsigned int i_a;
	unsigned int distance;
	struct synonym_array *array;
	struct synonym *synonym;
	struct result *result;

	if(l <= 4)
		return;

	i_a = st_char_index(s[0]);

	array = &thesaurus->index_a[i_a];
	n = array->n;
	for(i = 0; i < n; i++) {
		synonym = array->synonyms[i];
		if(synonym->word_length <= 4)
			continue;

		distance = edit_distance(s, l, synonym->word, synonym->word_length);
		if(distance <= 2) {
			result = xmalloc(sizeof(*result));
			result_init(result);
			result->synonym = synonym;
			transformations_or(&result->transformations,
				&synonym->transformations, t);
			result->index_a = 1;
			result->distance = distance;

			LIST_INSERT_BEFORE(result_list, result);
		}
	}
}

void
thesaurus_search_index_b(struct thesaurus *thesaurus,
	struct result *result_list,
	const char *s, unsigned int l,
	const struct transformations *t)
{
	unsigned int i;
	unsigned int n;
	unsigned int i_b;
	unsigned int distance;
	struct synonym_array *array;
	struct synonym *synonym;
	struct result *result;

	if(l <= 4)
		return;

	i_b = st_char_index(s[l - 1]);

	array = &thesaurus->index_b[i_b];
	n = array->n;
	for(i = 0; i < n; i++) {
		synonym = array->synonyms[i];
		if(synonym->word_length <= 4)
			continue;

		distance = edit_distance(s, l, synonym->word, synonym->word_length);
		if(distance <= 2) {
			result = xmalloc(sizeof(*result));
			result_init(result);
			result->synonym = synonym;
			transformations_or(&result->transformations,
				&synonym->transformations, t);
			result->index_b = 1;
			result->distance = distance;

			LIST_INSERT_BEFORE(result_list, result);
		}
	}
}

void
thesaurus_search_index_ab(struct thesaurus *thesaurus,
	struct result *result_list,
	const char *s, unsigned int l,
	const struct transformations *t)
{
	unsigned int i;
	unsigned int n;
	unsigned int i_a;
	unsigned int i_b;
	unsigned int distance;
	struct synonym_array *array;
	struct synonym *synonym;
	struct result *result;

	if(l <= 4)
		return;

	i_a = st_char_index(s[0]);
	i_b = st_char_index(s[l - 1]);

	array = &thesaurus->index_ab[i_a + 26 * i_b];
	n = array->n;
	for(i = 0; i < n; i++) {
		synonym = array->synonyms[i];
		if(synonym->word_length <= 4)
			continue;

#if 0
		/* XXX: Remember to put this in _a and _b too. */
		if(abs(l - synonym->word_length) >= 2)
			continue;
#endif

		distance = edit_distance(s, l,
			synonym->word, synonym->word_length);
		if(distance <= 2) {
			result = xmalloc(sizeof(*result));
			result_init(result);
			result->synonym = synonym;
			transformations_or(&result->transformations,
				&synonym->transformations, t);
			result->index_ab = 1;
			result->distance = distance;

			LIST_INSERT_BEFORE(result_list, result);
		}
	}
}

/* Don't use this. Way too slow. Outdated. */
/*
void
thesaurus_search_index(struct thesaurus *thesaurus,
	struct result *result_list,
	const char *s, unsigned int l)
{
	unsigned int i;
	unsigned int n;
	unsigned int distance;
	struct synonym_array *array;
	struct synonym *synonym;
	struct result *result;

	array = &thesaurus->index;
	n = array->n;
	for(i = 0; i < n; i++) {
		synonym = array->synonyms[i];
		distance = edit_distance(s, l, synonym->word, synonym->word_length);

		if(distance <= 2 && st_min(l, synonym->word_length) > 3) {
			result = xmalloc(sizeof(*result));
			result_init(result);
			result->synonym = synonym;
			result->index = 1;
			result->distance = distance;

			LIST_INSERT_BEFORE(result_list, result);
		}
	}
}
*/
