#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "libs/xmalloc.h"

#include "synonym-array.h"
#include "trie.h"

void
trie_init(struct trie *trie, char ch)
{
	assert(trie);

	trie->nchildren = 0;
	trie->children = NULL;
	trie->ch = ch;
	trie->synonyms = NULL;
}

void
trie_deinit(struct trie *trie)
{
	unsigned int i;

	assert(trie);

	if(trie->synonyms) {
		synonym_array_deinit(trie->synonyms, 0);
		free(trie->synonyms);
	}

	for(i = 0; i < trie->nchildren; i++)
		trie_deinit(&trie->children[i]);

	free(trie->children);
}

static int
st_trie_compare(const void *_a, const void *_b)
{
	const struct trie *a;
	const struct trie *b;

	a = _a;
	b = _b;

	if(a->ch < b->ch)
		return -1;

	if(a->ch > b->ch)
		return 1;

	return 0;
}

/* The binary search assumes that the children are stored in
 * order. The insertion algorithm inserts them in the input order,
 * so for the binsary search algorithm to work, the input must be
 * ordered. */
static struct trie *
st_trie_find_character_ordered(struct trie *trie, char ch)
{
	struct trie key;

	assert(trie);

	key.ch = ch;
	return bsearch(&key,
		trie->children, trie->nchildren,
		sizeof(struct trie), &st_trie_compare);
}

static struct trie *
st_trie_find_character(struct trie *trie, char ch)
{
	unsigned int i;

	assert(trie);

	for(i = 0; i < trie->nchildren; i++) {
		if(trie->children[i].ch == ch)
			return &trie->children[i];
	}

	return NULL;
}

/* Returns the inserted element */
struct trie *
trie_insert(struct trie *trie, const char *s, unsigned int l)
{
	unsigned int i;
	struct trie *cur, *new;

	assert(trie);
	assert(s);
	assert(l > 0);

	cur = trie;

	for(i = 0; i < l; i++) {
		/* Uses the slow, linear-search algorithm since the input
		 * may not be in order. */
		new = st_trie_find_character(cur, s[i]);
		if(new) {
			cur = new;
			continue;
		}

		cur->nchildren++;
		cur->children = xrealloc(cur->children,
			cur->nchildren * sizeof(struct trie));

		new = &cur->children[cur->nchildren - 1];
		trie_init(new, s[i]);

		cur = new;
	}

	return cur;
}

void
trie_sort(struct trie *trie)
{
	unsigned int i;

	assert(trie);

	/* Get out as quickly as possible. */
	if(!trie->nchildren)
		return;

	qsort(trie->children, trie->nchildren, sizeof(struct trie),
		&st_trie_compare);

	/* Recursively sort all children. */
	for(i = 0; i < trie->nchildren; i++)
		trie_sort(&trie->children[i]);
}

struct trie *
trie_search(struct trie *trie, const char *s, unsigned int l)
{
	unsigned int i;
	struct trie *cur, *new;

	assert(trie);
	assert(s);
	assert(l > 0);

	cur = trie;

	for(i = 0; i < l; i++) {
		/* Use the binary search algorithm. Note: trie_sort() MUST
		 * be called prior to any trie_search() call. */
		new = st_trie_find_character_ordered(cur, s[i]);
		if(!new)
			return NULL;

		cur = new;
	}

	return cur;
}
